#!/usr/bin/env python
#
# clearsettings.py - The ClearSettingsAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ClearSettingsAction` class, an action
which clears all settings from the :mod:`fsl.utils.settings` module.
"""


import fsl.utils.settings as fslsettings
import fsleyes.strings    as strings
from . import                base


class ClearSettingsAction(base.Action):
    """The ``ClearSettingsAction`` class is an action which clears all
    settings from the :mod:`fsl.utils.settings` module.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``ClearSettingsAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(
            self, overlayList, displayCtx, self.__clearSettings)
        self.__frame = frame


    def __clearSettings(self):
        """Ask the user to confirm, then clear  all settings stored by the
        :mod:`fsl.utils.settings` module.
        """

        import wx

        msg    = strings.messages[self, 'confirm']
        title  = strings.titles[  self, 'confirm']

        dlg = wx.MessageDialog(self.__frame,
                               message=msg,
                               caption=title,
                               style=(wx.YES_NO        |
                                      wx.NO_DEFAULT    |
                                      wx.CENTRE        |
                                      wx.ICON_WARNING))

        dlg.CentreOnParent()
        if dlg.ShowModal() == wx.ID_NO:
            return

        fslsettings.clear()



#!/usr/bin/env python
#
# resample.py - The ResampleAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ResampleAction` class, a FSLeyes action
which allows the user to resample an image to a different resolution.
"""

import collections

import          wx
import numpy as np

import fsleyes_widgets.floatspin as floatspin
import fsl.data.image            as fslimage
import fsl.utils.image.resample  as resample
import fsleyes.strings           as strings
import fsleyes.tooltips          as tooltips
import fsleyes.actions.base      as base


class ResampleAction(base.NeedOverlayAction):
    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``ResampleAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.NeedOverlayAction.__init__(
            self, overlayList, displayCtx, func=self.__resample)
        self.__frame = frame


    def __resample(self):
        """Called when this ``ResampleAction`` is invoked. Shows a
        ``ResampleDialog``, and then resamples the currently selected overlay.
        """

        ovl  = self.displayCtx.getSelectedOverlay()
        opts = self.displayCtx.getOpts(ovl)

        def refCandidate(o):
            return (isinstance(o, fslimage.Nifti) and
                    (o is not ovl)                and
                    not o.sameSpace(ovl))

        refs = [o for o in self.overlayList if refCandidate(o)]
        dlg  = ResampleDialog(
            self.__frame,
            title=ovl.name,
            shape=ovl.shape,
            pixdim=ovl.pixdim,
            refs=refs)

        if dlg.ShowModal() != wx.ID_OK:
            return

        newShape  = dlg.GetVoxels()
        interp    = dlg.GetInterpolation()
        origin    = dlg.GetOrigin()
        dtype     = dlg.GetDataType()
        smoothing = dlg.GetSmoothing()
        ref       = dlg.GetReference()
        allvols   = dlg.GetAllVolumes()
        interp    = {'nearest' : 0, 'linear' : 1, 'cubic' : 3}[interp]
        name      = '{}_resampled'.format(ovl.name)

        if allvols or ovl.ndim == 3: slc = None
        else:                        slc = opts.index()

        if allvols and ovl.ndim > 3:
            newShape = list(newShape) + list(ovl.shape[3:])

        kwargs = dict(sliceobj=slc,
                      dtype=dtype,
                      order=interp,
                      origin=origin,
                      smooth=smoothing)

        if ref is not None:
            resampled, xform = resample.resampleToReference(ovl, ref, **kwargs)
        else:
            resampled, xform = resample.resample(ovl, newShape, **kwargs)

        resampled = fslimage.Image(resampled,
                                   xform=xform,
                                   header=ovl.header,
                                   name=name)

        self.overlayList.append(resampled)


class ResampleDialog(wx.Dialog):
    """The ``ResampleDialog`` is used by the ``ResampleAction`` to prompt the
    user for a new resampled image shape. It contains controls allowing the
    user to select new voxel and pixdim values, and to select resampling
    options for interpolation, data type, and smoothing.
    """

    def __init__(self, parent, title, shape, pixdim, refs):
        """Create a ``ResampleDialog``.

        :arg parent: ``wx`` parent object
        :arg title:  Dialog title
        :arg shape:  The original image shape (a tuple of integers)
        :arg pixdim: The original image pixdims (a tuple of floats)
        :arg refs:   A sequence of :class:`.Image` objects which can
                     be selected as references.
        """

        wx.Dialog.__init__(self,
                           parent,
                           title=title,
                           style=wx.DEFAULT_DIALOG_STYLE)

        self.__oldShape   = tuple(shape)
        self.__oldPixdim  = tuple(pixdim)
        self.__references = collections.OrderedDict(
            [(strings.labels[self, 'noref'], None)] +
            [(r.name, r) for r in refs])

        self.__ok        = wx.Button(self, id=wx.ID_OK)
        self.__reset     = wx.Button(self)
        self.__cancel    = wx.Button(self, id=wx.ID_CANCEL)

        self.__ok    .SetLabel(strings.labels[self, 'ok'])
        self.__reset .SetLabel(strings.labels[self, 'reset'])
        self.__cancel.SetLabel(strings.labels[self, 'cancel'])

        voxargs = {'minValue'  : 1,
                   'maxValue'  : 9999,
                   'increment' : 1,
                   'width'     : 6}
        pixargs = {'minValue'  : 0.001,
                   'maxValue'  : 50,
                   'increment' : 0.5,
                   'width'     : 6}

        strvox = ['{:d}'   .format(p) for p in shape]
        strpix = ['{:0.2f}'.format(p) for p in pixdim]

        self.__refLabel     = wx.StaticText(self)
        self.__origVoxLabel = wx.StaticText(self)
        self.__origPixLabel = wx.StaticText(self)
        self.__voxLabel     = wx.StaticText(self)
        self.__pixLabel     = wx.StaticText(self)

        self.__refLabel    .SetLabel(strings.labels[self, 'reference'])
        self.__origVoxLabel.SetLabel(strings.labels[self, 'origVoxels'])
        self.__origPixLabel.SetLabel(strings.labels[self, 'origPixdims'])
        self.__voxLabel    .SetLabel(strings.labels[self, 'newVoxels'])
        self.__pixLabel    .SetLabel(strings.labels[self, 'newPixdims'])

        self.__origVoxx = wx.StaticText(self, label=strvox[0])
        self.__origVoxy = wx.StaticText(self, label=strvox[1])
        self.__origVoxz = wx.StaticText(self, label=strvox[2])
        self.__origPixx = wx.StaticText(self, label=strpix[0])
        self.__origPixy = wx.StaticText(self, label=strpix[1])
        self.__origPixz = wx.StaticText(self, label=strpix[2])

        self.__refChoice = wx.Choice(self,
                                     choices=list(self.__references.keys()))
        self.__refChoice.Enable(len(self.__references) > 1)
        self.__refChoice.SetSelection(0)

        self.__voxx = floatspin.FloatSpinCtrl(self, value=shape[ 0], **voxargs)
        self.__voxy = floatspin.FloatSpinCtrl(self, value=shape[ 1], **voxargs)
        self.__voxz = floatspin.FloatSpinCtrl(self, value=shape[ 2], **voxargs)
        self.__pixx = floatspin.FloatSpinCtrl(self, value=pixdim[0], **pixargs)
        self.__pixy = floatspin.FloatSpinCtrl(self, value=pixdim[1], **pixargs)
        self.__pixz = floatspin.FloatSpinCtrl(self, value=pixdim[2], **pixargs)

        self.__originChoices = ['centre', 'corner']
        self.__interpChoices = ['linear', 'nearest', 'cubic']
        self.__dtypeChoices  = [('float',  np.float32),
                                ('uchar',  np.uint8),
                                ('sshort', np.int16),
                                ('sint',   np.int32),
                                ('double', np.float64)]

        self.__originLabels  = [strings.labels[self, c]
                                for c in self.__originChoices]
        self.__interpLabels  = [strings.labels[self, c]
                                for c in self.__interpChoices]
        self.__dtypeLabels   = [strings.labels[self, c[0]]
                                for c in self.__dtypeChoices]

        self.__originLabel = wx.StaticText(self)
        self.__interpLabel = wx.StaticText(self)
        self.__dtypeLabel  = wx.StaticText(self)
        self.__smoothLabel = wx.StaticText(self)
        self.__allVolLabel = wx.StaticText(self)
        self.__origin      = wx.Choice(self, choices=self.__originLabels)
        self.__interp      = wx.Choice(self, choices=self.__interpLabels)
        self.__dtype       = wx.Choice(self, choices=self.__dtypeLabels)
        self.__smooth      = wx.CheckBox(self)
        self.__allVolumes  = wx.CheckBox(self)

        if len(shape) <= 3:
            self.__allVolumes .Disable()
            self.__allVolLabel.Disable()

        self.__origin    .SetSelection(0)
        self.__interp    .SetSelection(0)
        self.__dtype     .SetSelection(0)
        self.__smooth    .SetValue(True)
        self.__allVolumes.SetValue(True)

        self.__originLabel.SetLabel(strings.labels[self, 'origin'])
        self.__interpLabel.SetLabel(strings.labels[self, 'interpolation'])
        self.__dtypeLabel .SetLabel(strings.labels[self, 'dtype'])
        self.__smoothLabel.SetLabel(strings.labels[self, 'smoothing'])
        self.__allVolLabel.SetLabel(strings.labels[self, 'allVolumes'])

        self.__refLabel.SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'reference']))
        self.__refChoice.SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'reference']))
        self.__originLabel.SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'origin']))
        self.__origin.SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'origin']))
        self.__interp     .SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'interpolation']))
        self.__interpLabel.SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'interpolation']))
        self.__dtype      .SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'dtype']))
        self.__dtypeLabel .SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'dtype']))
        self.__smooth     .SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'smoothing']))
        self.__smoothLabel.SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'smoothing']))
        self.__allVolumes .SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'allVolumes']))
        self.__allVolLabel.SetToolTip(
            wx.ToolTip(tooltips.misc[self, 'allVolumes']))

        self.__refSizer    = wx.BoxSizer(wx.HORIZONTAL)
        self.__labelSizer  = wx.BoxSizer(wx.HORIZONTAL)
        self.__xrowSizer   = wx.BoxSizer(wx.HORIZONTAL)
        self.__yrowSizer   = wx.BoxSizer(wx.HORIZONTAL)
        self.__zrowSizer   = wx.BoxSizer(wx.HORIZONTAL)
        self.__originSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__interpSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__dtypeSizer  = wx.BoxSizer(wx.HORIZONTAL)
        self.__smoothSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__allVolSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__btnSizer    = wx.BoxSizer(wx.HORIZONTAL)
        self.__mainSizer   = wx.BoxSizer(wx.VERTICAL)

        self.__refSizer.Add((10, 10),         flag=wx.EXPAND)
        self.__refSizer.Add(self.__refLabel,  flag=wx.EXPAND, proportion=1)
        self.__refSizer.Add((10, 10),         flag=wx.EXPAND)
        self.__refSizer.Add(self.__refChoice, flag=wx.EXPAND, proportion=1)
        self.__refSizer.Add((10, 10),         flag=wx.EXPAND)

        self.__labelSizer.Add((10, 10),            flag=wx.EXPAND)
        self.__labelSizer.Add(self.__origVoxLabel, flag=wx.EXPAND,
                              proportion=1)
        self.__labelSizer.Add((10, 10),            flag=wx.EXPAND)
        self.__labelSizer.Add(self.__origPixLabel, flag=wx.EXPAND,
                              proportion=1)
        self.__labelSizer.Add((10, 10),            flag=wx.EXPAND)
        self.__labelSizer.Add(self.__voxLabel,     flag=wx.EXPAND,
                              proportion=1)
        self.__labelSizer.Add((10, 10),            flag=wx.EXPAND)
        self.__labelSizer.Add(self.__pixLabel,     flag=wx.EXPAND,
                              proportion=1)
        self.__labelSizer.Add((10, 10),            flag=wx.EXPAND)

        self.__xrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__xrowSizer.Add(self.__origVoxx, flag=wx.EXPAND, proportion=1)
        self.__xrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__xrowSizer.Add(self.__origPixx, flag=wx.EXPAND, proportion=1)
        self.__xrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__xrowSizer.Add(self.__voxx,     flag=wx.EXPAND, proportion=1)
        self.__xrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__xrowSizer.Add(self.__pixx,     flag=wx.EXPAND, proportion=1)
        self.__xrowSizer.Add((10, 10),        flag=wx.EXPAND)

        self.__yrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__yrowSizer.Add(self.__origVoxy, flag=wx.EXPAND, proportion=1)
        self.__yrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__yrowSizer.Add(self.__origPixy, flag=wx.EXPAND, proportion=1)
        self.__yrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__yrowSizer.Add(self.__voxy,     flag=wx.EXPAND, proportion=1)
        self.__yrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__yrowSizer.Add(self.__pixy,     flag=wx.EXPAND, proportion=1)
        self.__yrowSizer.Add((10, 10),        flag=wx.EXPAND)

        self.__zrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__zrowSizer.Add(self.__origVoxz, flag=wx.EXPAND, proportion=1)
        self.__zrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__zrowSizer.Add(self.__origPixz, flag=wx.EXPAND, proportion=1)
        self.__zrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__zrowSizer.Add(self.__voxz,     flag=wx.EXPAND, proportion=1)
        self.__zrowSizer.Add((10, 10),        flag=wx.EXPAND)
        self.__zrowSizer.Add(self.__pixz,     flag=wx.EXPAND, proportion=1)
        self.__zrowSizer.Add((10, 10),        flag=wx.EXPAND)

        self.__originSizer.Add((50, 1),            flag=wx.EXPAND)
        self.__originSizer.Add(self.__originLabel, flag=wx.EXPAND)
        self.__originSizer.Add((10, 1),            flag=wx.EXPAND)
        self.__originSizer.Add(self.__origin,      flag=wx.EXPAND)
        self.__originSizer.Add((10, 1),            flag=wx.EXPAND,
                               proportion=1)

        self.__interpSizer.Add((50, 1),            flag=wx.EXPAND)
        self.__interpSizer.Add(self.__interpLabel, flag=wx.EXPAND)
        self.__interpSizer.Add((10, 1),            flag=wx.EXPAND)
        self.__interpSizer.Add(self.__interp,      flag=wx.EXPAND)
        self.__interpSizer.Add((10, 1),            flag=wx.EXPAND,
                               proportion=1)

        self.__dtypeSizer.Add((50, 1),           flag=wx.EXPAND)
        self.__dtypeSizer.Add(self.__dtypeLabel, flag=wx.EXPAND)
        self.__dtypeSizer.Add((10, 1),           flag=wx.EXPAND)
        self.__dtypeSizer.Add(self.__dtype,      flag=wx.EXPAND)
        self.__dtypeSizer.Add((10, 1),           flag=wx.EXPAND,
                               proportion=1)

        self.__smoothSizer.Add((50, 1),            flag=wx.EXPAND)
        self.__smoothSizer.Add(self.__smoothLabel, flag=wx.EXPAND)
        self.__smoothSizer.Add((10, 1),            flag=wx.EXPAND)
        self.__smoothSizer.Add(self.__smooth,      flag=wx.EXPAND)
        self.__smoothSizer.Add((10, 1),            flag=wx.EXPAND,
                               proportion=1)

        self.__allVolSizer.Add((50, 1),            flag=wx.EXPAND)
        self.__allVolSizer.Add(self.__allVolLabel, flag=wx.EXPAND)
        self.__allVolSizer.Add((10, 1),            flag=wx.EXPAND)
        self.__allVolSizer.Add(self.__allVolumes,  flag=wx.EXPAND)
        self.__allVolSizer.Add((10, 1),            flag=wx.EXPAND,
                               proportion=1)

        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND, proportion=1)
        self.__btnSizer.Add(self.__ok,     flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND)
        self.__btnSizer.Add(self.__reset,  flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND)
        self.__btnSizer.Add(self.__cancel, flag=wx.EXPAND)
        self.__btnSizer.Add((10, 1),       flag=wx.EXPAND, proportion=1)

        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__refSizer,    flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__labelSizer,  flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__xrowSizer,   flag=wx.EXPAND)
        self.__mainSizer.Add(self.__yrowSizer,   flag=wx.EXPAND)
        self.__mainSizer.Add(self.__zrowSizer,   flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__originSizer, flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__interpSizer, flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__dtypeSizer,  flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__smoothSizer, flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__allVolSizer, flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)
        self.__mainSizer.Add(self.__btnSizer,    flag=wx.EXPAND)
        self.__mainSizer.Add((10, 10),           flag=wx.EXPAND)

        self.SetSizer(self.__mainSizer)

        self.__ok       .Bind(wx.EVT_BUTTON,           self.__onOk)
        self.__reset    .Bind(wx.EVT_BUTTON,           self.__onReset)
        self.__cancel   .Bind(wx.EVT_BUTTON,           self.__onCancel)
        self.__refChoice.Bind(wx.EVT_CHOICE,           self.__onRef)
        self.__voxx     .Bind(floatspin.EVT_FLOATSPIN, self.__onVoxel)
        self.__voxy     .Bind(floatspin.EVT_FLOATSPIN, self.__onVoxel)
        self.__voxz     .Bind(floatspin.EVT_FLOATSPIN, self.__onVoxel)
        self.__pixx     .Bind(floatspin.EVT_FLOATSPIN, self.__onPixdim)
        self.__pixy     .Bind(floatspin.EVT_FLOATSPIN, self.__onPixdim)
        self.__pixz     .Bind(floatspin.EVT_FLOATSPIN, self.__onPixdim)

        self.__ok.SetDefault()

        self.Layout()
        self.Fit()
        self.CentreOnParent()


    @property
    def okButton(self):
        """Returns a reference to the OK button. """
        return self.__ok


    @property
    def resetButton(self):
        """Returns a reference to the reset button. """
        return self.__reset


    @property
    def cancelButton(self):
        """Returns a reference to the cancel button. """
        return self.__cancel


    @property
    def refCtrl(self):
        """Returns a reference to the reference image dropdown. """
        return self.__refChoice


    @property
    def voxXCtrl(self):
        """Returns a reference to the X voxel coordinate spin control. """
        return self.__voxx


    @property
    def voxYCtrl(self):
        """Returns a reference to the Y voxel coordinate spin control. """
        return self.__voxy


    @property
    def voxZCtrl(self):
        """Returns a reference to the Z voxel coordinate spin control. """
        return self.__voxz


    @property
    def pixXCtrl(self):
        """Returns a reference to the X pixdim spin control. """
        return self.__pixx


    @property
    def pixYCtrl(self):
        """Returns a reference to the Y pixdim spin control. """
        return self.__pixy


    @property
    def pixZCtrl(self):
        """Returns a reference to the Z pixdim spin control. """
        return self.__pixz


    @property
    def originCtrl(self):
        """Returns a reference to the origin dropdown box. """
        return self.__origin


    @property
    def interpCtrl(self):
        """Returns a reference to the interpolation dropdown box. """
        return self.__interp


    @property
    def dtypeCtrl(self):
        """Returns a reference to the data type dropdown box. """
        return self.__dtype


    @property
    def smoothCtrl(self):
        """Returns a reference to the smoothing checkbox. """
        return self.__smooth


    @property
    def allVolumesCtrl(self):
        """Returns a reference to the all volumes checkbox. """
        return self.__allVolumes


    def __onRef(self, ev):
        """Called when the user changes the reference image. Enables/
        disables the voxel/pixdim controls as needed.
        """
        ref   = self.__refChoice.GetSelection()
        ref   = self.__refChoice.GetString(ref)
        ref   = self.__references[ref]
        noref = ref is None

        self.__voxx  .Enable(noref)
        self.__voxy  .Enable(noref)
        self.__voxz  .Enable(noref)
        self.__pixx  .Enable(noref)
        self.__pixy  .Enable(noref)
        self.__pixz  .Enable(noref)
        self.__origin.Enable(noref)


    def __onVoxel(self, ev):
        """Called when the user changes a voxel value. Updates the pixdim
        values accordingly.
        """

        newpix = self.__derivePixdims()

        self.__pixx.SetValue(newpix[0])
        self.__pixy.SetValue(newpix[1])
        self.__pixz.SetValue(newpix[2])


    def __onPixdim(self, ev):
        """Called when the user changes a pixdim value. Updates the voxel
        values accordingly.
        """

        newvox = self.__deriveVoxels()

        self.__voxx.SetValue(newvox[0])
        self.__voxy.SetValue(newvox[1])
        self.__voxz.SetValue(newvox[2])


    def __onOk(self, ev):
        """Called when the ok button is pushed. Closes the dialog. """
        self.EndModal(wx.ID_OK)


    def __onReset(self, ev):
        """Called when the reset button is pushed. Resets the shape and pixdims
        to their original values.
        """
        self.__voxx.SetValue(self.__oldShape[ 0])
        self.__voxy.SetValue(self.__oldShape[ 1])
        self.__voxz.SetValue(self.__oldShape[ 2])
        self.__pixx.SetValue(self.__oldPixdim[0])
        self.__pixy.SetValue(self.__oldPixdim[1])
        self.__pixz.SetValue(self.__oldPixdim[2])

        self.__interp.SetSelection(0)
        self.__dtype .SetSelection(0)


    def __onCancel(self, ev):
        """Called when the cancel button is pushed. Closes the dialog. """
        self.EndModal(wx.ID_CANCEL)


    def GetVoxels(self):
        """Returns the current voxel values. """
        return (self.__voxx.GetValue(),
                self.__voxy.GetValue(),
                self.__voxz.GetValue())


    def GetInterpolation(self):
        """Returns the currently selected interpolation setting, either
        ``'nearest'``, ``'linear'``, or ``'cubic'``.
        """
        choice = self.__interp.GetSelection()
        return self.__interpChoices[choice]


    def GetOrigin(self):
        """Returns the currently selected origin setting, either
        ``'corner'``, or ``'centre'``.
        """
        choice = self.__origin.GetSelection()
        return self.__originChoices[choice]


    def GetDataType(self):
        """Returns the currently selected data type setting as a
        ``numpy.dtype``, one of ``uint8``, ``int16``, ``int32``, ``float32``,
        or ``float64``.
        """
        choice = self.__dtype.GetSelection()
        return self.__dtypeChoices[choice][1]


    def GetSmoothing(self):
        """Returns the currently selected smoothing setting, either
        ``True``, or ``False``.
        """
        return self.__smooth.GetValue()


    def GetAllVolumes(self):
        """Returns ``True``, or ``False``, indicating whether all
        volumes of an image with more than three dimensions should be
        resampled.
        """
        return self.__allVolumes.GetValue()


    def GetPixdims(self):
        """Returns the current pixdim values. """
        return (self.__pixx.GetValue(),
                self.__pixy.GetValue(),
                self.__pixz.GetValue())


    def GetReference(self):
        """Return the current reference image value. """
        ref = self.__refChoice.GetSelection()
        ref = self.__refChoice.GetString(ref)
        return self.__references[ref]


    def __derivePixdims(self):
        """Derives new pixdim values from the current voxel values. """
        olds = self.__oldShape
        oldp = self.__oldPixdim
        news = self.GetVoxels()
        fac  = [o / float(n) for o, n in zip(olds, news)]
        newp = [p * f        for p, f in zip(oldp, fac)]

        return newp


    def __deriveVoxels(self):
        """Derives new voxel values from the current pixdim values. """
        olds = self.__oldShape
        oldp = self.__oldPixdim
        newp = self.GetPixdims()
        fac  = [o / float(n) for o, n in zip(oldp, newp)]
        news = [p * f        for p, f in zip(olds, fac)]

        return news



#!/usr/bin/env python


import                                  wx
import wx.lib.agw.aui                as aui
import wx.lib.agw.aui.framemanager   as auifm
import fsleyes_widgets               as fwidgets
import fsleyes.controls.controlpanel as ctrlpanel


class MyAuiFloatingFrame(auifm.AuiFloatingFrame):
    """Here I am monkey patching the
    ``wx.agw.aui.framemanager.AuiFloatingFrame.__init__`` method.

    I am doing this because I have observed some strange behaviour when running
    a remote instance of this application over an SSH/X11 session, with the X11
    server (i.e. the local machine) running in OS X. When a combobox is embedded
    in a floating frame (either a pane or a toolbar), its dropdown list appears
    underneath the frame, meaning that the user is unable to actually select any
    items from the list!

    I have only seen this behaviour when using XQuartz on macOS.

    Ultimately, this appears to be caused by the ``wx.FRAME_TOOL_WINDOW``
    style, as passed to the ``wx.MiniFrame`` constructor (from which the
    ``AuiFloatingFrame`` class derives). Removing this style flag fixes the
    problem, so this is exactly what I'm doing. I haven't looked any deeper
    into the situation.


    This class also overrieds the ``SetPaneWindow`` method, because under gtk3,
    the maximum size if a frame musr be set.
    """

    def __init__(self, *args, **kwargs):
        """My new constructor, which makes sure that the ``FRAME_TOOL_WINDOW``
        style is not passed through to the ``AuiFloatingFrame`` constructor
        """

        if 'style' in kwargs:
            style = kwargs['style']

        # This is the default style, as defined
        # in the AuiFloatingFrame constructor
        else:
            style = (wx.FRAME_TOOL_WINDOW     |
                     wx.FRAME_FLOAT_ON_PARENT |
                     wx.FRAME_NO_TASKBAR      |
                     wx.CLIP_CHILDREN)

        if fwidgets.inSSHSession():
            style &= ~wx.FRAME_TOOL_WINDOW

        kwargs['style'] = style

        super().__init__(*args, **kwargs)


    def SetPaneWindow(self, pane):
        """Make sure that floated toolbars are sized correctly.
        """
        super().SetPaneWindow(pane)
        if isinstance(pane.window, ctrlpanel.ControlToolBar):
            size = self.GetBestSize()
            self.SetMaxSize(size)


def _AuiDockingGuide_init(self, *args, **kwargs):
    """I am also monkey-patching the
    ``wx.lib.agw.aui.AuiDockingGuide.__init__`` method, because in this
    instance, when running over SSH/X11, the ``wx.FRAME_TOOL_WINDOW`` style
    seems to result in the docking guide frames being given title bars, which
    is quite undesirable.

    I cannot patch the entire class in the aui package, because it is used
    as part of a class hierarchy. So I am just patching the method.
    """

    if 'style' in kwargs:
        style = kwargs['style']

    # This is the default style, as defined
    # in the AuiDockingGuide constructor
    else:
        style = (wx.FRAME_TOOL_WINDOW |
                 wx.FRAME_STAY_ON_TOP |
                 wx.FRAME_NO_TASKBAR  |
                 wx.NO_BORDER)

    if fwidgets.inSSHSession():
        style &= ~wx.FRAME_TOOL_WINDOW

    kwargs['style'] = style

    _AuiDockingGuide_real_init(self, *args, **kwargs)


if fwidgets.inSSHSession():
    aui  .AuiFloatingFrame       = MyAuiFloatingFrame
    auifm.AuiFloatingFrame       = MyAuiFloatingFrame
    _AuiDockingGuide_real_init   = aui.AuiDockingGuide.__init__
    aui.AuiDockingGuide.__init__ = _AuiDockingGuide_init



#!/usr/bin/env python
#
# scene3dopts.py - The Scene3DOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Scene3DOpts` class, which is used
by :class:`.Scene3DPanel` instances for managing their display settings.
"""


import copy

from . import sceneopts
from . import canvasopts


class Scene3DOpts(sceneopts.SceneOpts):
    """The ``Scene3DOpts`` class contains display settings for the
    :class:`.Scene3DPanel` class.

    All of the properties in the ``Scene3DOpts`` class are defined in the
    :class:`.Scene3DCanvasOpts` class - see its documentation for more
    details.
    """


    showLegend    = copy.copy(canvasopts.Scene3DCanvasOpts.showLegend)
    light         = copy.copy(canvasopts.Scene3DCanvasOpts.light)
    lightPos      = copy.copy(canvasopts.Scene3DCanvasOpts.lightPos)
    lightDistance = copy.copy(canvasopts.Scene3DCanvasOpts.lightDistance)
    showLight     = copy.copy(canvasopts.Scene3DCanvasOpts.showLight)
    offset        = copy.copy(canvasopts.Scene3DCanvasOpts.offset)
    rotation      = copy.copy(canvasopts.Scene3DCanvasOpts.rotation)


    def __init__(self, *args, **kwargs):
        """Create a ``Scene3DCanvasOpts`` instance. All arguments are passed
        through to the :class:`.SceneOpts` constructor.
        """
        self.setAttribute('zoom',     'minval',  75)
        self.setAttribute('zoom',     'default', 75)
        self.setAttribute('zoom',     'maxval',  5000)
        self.setAttribute('bgColour', 'default', (0.6, 0.6, 0.753, 1.0))
        self.setAttribute('fgColour', 'default', (0.0, 1.0, 0.0,   1.0))

        for ax in range(3):
            self.lightPos.setLimits(ax, -180, 180)

        self.zoom     = 75
        self.bgColour = (0.6, 0.6, 0.753)
        self.fgColour = (0,   1,   0)

        sceneopts.SceneOpts.__init__(self, *args, **kwargs)



#!/usr/bin/env python
#
# gltractogram_funcs.py - GL21 functions for drawing tractogram overlays.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module comtains functions for drawing tractogram overlays with OpenGL
2.1. These functions are used by :class:`.GLTractogram` instances.
"""

import itertools as it
import OpenGL.GL as gl
import numpy     as np

import fsl.transform.affine  as affine
import fsleyes.gl.routines   as glroutines
import fsleyes.gl.extensions as glexts
import fsleyes.gl.shaders    as shaders


def compileShaders(self):
    """Called by :meth:`.GLTractogram.compileShaders`.
    Compiles shader programs.
    """

    # See comments in gl33.gltractogram_funcs.compileShaders
    vsrc       = shaders.getVertexShader(  'gltractogram')
    orientfsrc = shaders.getFragmentShader('gltractogram_orient')
    vdatafsrc  = shaders.getFragmentShader('gltractogram_vertex_data')
    idatafsrc  = shaders.getFragmentShader('gltractogram_image_data')

    colourSources = {
        'orientation' : orientfsrc,
        'vertexData'  : vdatafsrc,
        'imageData'   : idatafsrc,
    }

    colourModes = ['orientation', 'vertexData', 'imageData']
    clipModes   = ['none',        'vertexData', 'imageData']
    kwa         = {'resourceName' : f'GLTractogram_{id(self)}',
                   'shared'       : ['vertex']}

    for colourMode, clipMode in it.product(colourModes, clipModes):

        fsrc   = colourSources[colourMode]
        consts = {
            'colourMode' : colourMode,
            'clipMode'   : clipMode,
            'lighting'   : False,
            'twod'       : not self.threedee,
        }
        shader = shaders.GLSLShader(vsrc, fsrc, constants=consts, **kwa)

        self.shaders[colourMode][clipMode].append(shader)


def draw2D(self, canvas, mvp):
    """Called by :class:`.GLTractogram.draw2D`. """

    opts       = self.opts
    colourMode = opts.effectiveColourMode
    clipMode   = opts.effectiveClipMode
    res        = max((opts.resolution, 3))
    shader     = self.shaders[colourMode][clipMode][0]

    # each vertex is drawn as a circle,
    # using instanced rendering.
    vertices         = glroutines.unitCircle(res)
    scales           = self.normalisedLineWidth(canvas, mvp)
    vertices[:, :2] *= scales[:2]

    gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

    with shader.loaded(), shader.loadedAtts():
        shader.set(   'MVP',          mvp)
        shader.setAtt('circleVertex', vertices)
        glexts.glDrawArraysInstanced(gl.GL_TRIANGLE_FAN,
                                     0,
                                     len(vertices),
                                     len(self.vertices))


def draw3D(self, canvas, xform=None):
    """Called by :class:`.GLTractogram.draw3D`. """
    opts       = self.opts
    ovl        = self.overlay
    display    = self.display
    colourMode = opts.effectiveColourMode
    clipMode   = opts.effectiveClipMode
    vertXform  = ovl.affine
    mvp        = canvas.mvpMatrix
    mv         = canvas.viewMatrix
    lineWidth  = opts.lineWidth
    offsets    = self.offsets
    counts     = self.counts
    nstrms     = len(offsets)
    shader     = self.shaders[colourMode][clipMode][0]

    if xform is None: xform = vertXform
    else:             xform = affine.concat(xform, vertXform)

    mvp = affine.concat(mvp, xform)
    mv  = affine.concat(mv,  xform)

    with shader.loaded(), shader.loadedAtts():
        shader.set('MVP', mvp)
        # See comments in gl33.gltractogram_funcs.draw3D
        with glroutines.enabled(gl.GL_CULL_FACE):
            # we don't implement line width in gl21 - we
            # would need to use instanced rendering to
            # draw each line segment as a rectangle.
            gl.glLineWidth(lineWidth)
            gl.glCullFace(gl.GL_BACK)
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
            if display.alpha < 100 or opts.modulateAlpha:
                gl.glMultiDrawArrays(gl.GL_LINE_STRIP, offsets, counts, nstrms)
            with glroutines.enabled(gl.GL_DEPTH_TEST):
                gl.glMultiDrawArrays(gl.GL_LINE_STRIP, offsets, counts, nstrms)



#!/usr/bin/env python
#
# gllabel_funcs.py - OpenGL 1.4 functions used by the GLLabel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLLabel`
class to render :class:`.Image` overlays in an OpenGL 1.4 compatible manner.


Rendering of a ``GLLabel`` is very similar to that of a :class:`.GLVolume`,
with the exception that a different fragment program (``gllabel``) is used.
Therefore, the ``preDraw``, ``draw``, ``drawAll`` and ``postDraw`` functions
defined in the :mod:`.gl14.glvolume_funcs` are re-used by this module.
"""


import fsleyes.gl.shaders as shaders
from . import                glvolume_funcs


def init(self):
    """Calls the :func:`compileShaders` and :func:`updateShaderState`
    functions.
    """

    self.shader = None

    compileShaders(   self)
    updateShaderState(self)


def destroy(self):
    """Deletes handles to the vertex/fragment shader programs. """
    if self.shader is not None:
        self.shader.destroy()
    self.shader = None


def compileShaders(self):
    """Loads vertex and fragment shader program source, and creates an
    :class:`.ARBPShader` instance.
    """
    if self.shader is not None:
        self.shader.destroy()

    vertSrc  = shaders.getVertexShader(  'glvolume')
    fragSrc  = shaders.getFragmentShader('gllabel')
    textures = {
        'imageTexture' : 0,
        'lutTexture'   : 1,
    }

    self.shader = shaders.ARBPShader(vertSrc, fragSrc, textures)


def updateShaderState(self):
    """Updates all shader program variables. """

    if not self.ready():
        return

    opts = self.opts

    voxValXform  = self.imageTexture.voxValXform
    voxValXform  = [voxValXform[0, 0], voxValXform[0, 3], 0, 0]
    invNumLabels = 1.0 / (opts.lut.max() + 1)

    with self.shader.loaded():
        self.shader.setFragParam('voxValXform',  voxValXform)
        self.shader.setFragParam('invNumLabels', [invNumLabels, 0, 0, 0])

    return True


def draw2D(self, *args, **kwargs):
    """Draws the label overlay in 2D. See :meth:`.GLObject.draw2D`. """
    with self.shader.loaded():
        glvolume_funcs.draw2D(self, *args, **kwargs)


def drawAll(self, *args, **kwargs):
    """Draws the label overlay in 2D. See :meth:`.GLObject.draw2D`. """
    with self.shader.loaded():
        glvolume_funcs.drawAll(self, *args, **kwargs)



#!/usr/bin/env python
#
# saveoverlay.py - Save the currently selected overlay.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`SaveOverlayAction`, which allows the user
to save the currently selected overlay. A couple of standalone functions are
defined in this module, which do the real work:

.. autosummary::
   :nosignatures:

   saveOverlay
   doSave
   checkOverlaySaveState
"""


import logging

import                                 os
import os.path                      as op

import fsl.utils.settings           as fslsettings
import fsl.data.image               as fslimage
import fsleyes_widgets.utils.status as status
import fsleyes.strings              as strings
from . import                          base


log = logging.getLogger(__name__)


class SaveOverlayAction(base.Action):
    """The ``SaveOverlayAction`` allows the user to save the currently
    selected overlay, if it has been edited, or only exists in memory.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``SaveOverlayAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.Action.__init__(self, overlayList, displayCtx, self.__saveOverlay)
        self.__name       = '{}_{}'.format(type(self).__name__, id(self))
        self.__registered = None

        displayCtx .addListener('selectedOverlay',
                                self.__name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.__name,
                                self.__selectedOverlayChanged)

        self.__selectedOverlayChanged()


    def destroy(self):
        """Removes listeners from the :class:`.DisplayContext` and
        :class:`.OverlayList`, and calls :meth:`.Action.destroy`.
        """
        self.displayCtx .removeListener('selectedOverlay', self.__name)
        self.overlayList.removeListener('overlays',        self.__name)

        if self.__registered is not None:
            self.__registered.deregister(self.__name, 'saveState')
            self.__registered = None

        base.Action.destroy(self)


    def __selectedOverlayChanged(self, *a):
        """Called when the selected overlay, or overlay list changes.

        If the overlay is a :class:`.Image`, and it has unsaved changes,
        this action is enabled; otherwise it is disabled.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        # TODO  Support for other overlay types

        self.enabled = ((overlay is not None)               and
                        isinstance(overlay, fslimage.Image) and
                        (not overlay.saveState))

        if self.__registered is not None:
            self.__registered.deregister(self.__name, 'saveState')
            self.__registered = None

        # Register a listener on the saved property
        # of the currently selected image, so we can
        # enable the save action when the image
        # becomes 'unsaved', and vice versa.
        if self.enabled:
            self.__registered = overlay
            overlay.register(self.__name,
                             self.__overlaySaveStateChanged,
                             'saveState')


    def __overlaySaveStateChanged(self, *a):
        """Called when the :attr:`.Image.saved` property of the currently
        selected overlay changes. Enables/disables this ``SaveOverlayAction``
        accordingly.

        This is only applicable if the current overlay is a :class:`.Image` -
        see the :meth:`__selectedOverlayChanged` method.
        """

        overlay = self.__registered
        self.enabled = (overlay is not None) and (not overlay.saveState)


    def __saveOverlay(self):
        """Called when this :class:`.Action` is executed. Calls
        :func:`saveOverlay` with the currently selected overlay.
        """

        overlay = self.__registered
        if overlay is not None:
            display = self.displayCtx.getDisplay(overlay)
            saveOverlay(overlay, display)


def saveOverlay(overlay, display=None):
    """Saves the currently selected overlay (only if it is a :class:`.Image`),
    by a call to :meth:`.Image.save`. If a ``display`` is provided, the
    :attr:`.Display.name` may be updated to match the new overlay file name.

    :arg overlay: The :class:`.Image` overlay to save
    :arg display: The :class:`.Display` instance associated with the overlay.
    """


    import wx

    # TODO support for other overlay types
    if not isinstance(overlay, fslimage.Image):
        raise RuntimeError('Non-volumetric types not supported yet')

    # If this image has been loaded from a file,
    # ask the user whether they want to overwrite
    # that file, or save the image to a new file.
    #
    if overlay.dataSource is not None:

        # If the data source is not nifti (e.g.
        # mgz), we are not going to overwrite it,
        # so we don't ask.
        if fslimage.looksLikeImage(overlay.dataSource):

            msg   = strings.messages['SaveOverlayAction.overwrite'].format(
                overlay.dataSource)
            title = strings.titles[  'SaveOverlayAction.overwrite'].format(
                overlay.dataSource)

            dlg = wx.MessageDialog(
                wx.GetTopLevelWindows()[0],
                message=msg,
                caption=title,
                style=(wx.ICON_WARNING  |
                       wx.YES_NO        |
                       wx.CANCEL        |
                       wx.NO_DEFAULT))
            dlg.SetYesNoCancelLabels(
                strings.labels['SaveOverlayAction.overwrite'],
                strings.labels['SaveOverlayAction.saveNew'],
                strings.labels['SaveOverlayAction.cancel'])

            response = dlg.ShowModal()

            # Cancel == cancel the save
            # Yes    == overwrite the existing file
            # No     == save to a new file (prompt the user for the file name)
            if response == wx.ID_CANCEL:
                return

            if response == wx.ID_YES:
                doSave(overlay)
                return

        fromDir  = op.dirname(overlay.dataSource)
        filename = fslimage.removeExt(op.basename(overlay.dataSource))
        filename = '{}_copy'.format(filename)
    else:
        fromDir  = fslsettings.read('loadSaveOverlayDir', os.getcwd())

        if display is not None: filename = display.name
        else:                   filename = overlay.name

    filename = filename.replace('/',  '_')
    filename = filename.replace('\\', '_')

    # Ask the user where they
    # want to save the image
    msg = strings.titles['SaveOverlayAction.saveFile']
    dlg = wx.FileDialog(wx.GetApp().GetTopWindow(),
                        message=msg,
                        defaultDir=fromDir,
                        defaultFile=filename,
                        style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

    if dlg.ShowModal() != wx.ID_OK:
        return

    # Make sure that the user chose a supported
    # extension. If not, use the default extension.
    savePath       = dlg.GetPath()
    prefix, suffix = fslimage.splitExt(savePath)

    if suffix == '':
        savePath = '{}{}'.format(prefix, fslimage.defaultExt())

    oldPath = overlay.dataSource
    saveDir = op.dirname(savePath)

    if doSave(overlay, savePath):

        # Cache the save directory for next time.
        fslsettings.write('loadSaveOverlayDir', saveDir)

        # If image was in memory, or its old
        # name equalled the old datasource
        # base name, update its name.
        if oldPath is None or \
           fslimage.removeExt(op.basename(oldPath)) == overlay.name:

            overlay.name = fslimage.removeExt(op.basename(savePath))

            if display is not None:
                display.name = overlay.name


def doSave(overlay, path=None):
    """Called by :func:`saveOverlay`.  Tries to save the given ``overlay`` to
    the given ``path``, and shows an error message if something goes wrong.
    Returns ``True`` if the save was successful, ``False`` otherwise.
    """

    emsg   = strings.messages['SaveOverlayAction.saveError'].format(path)
    etitle = strings.titles[  'SaveOverlayAction.saveError']

    with status.reportIfError(msg=emsg, title=etitle, raiseError=False):
        overlay.save(path)
        return True

    return False


def checkOverlaySaveState(overlayList, displayCtx):
    """Returns ``True`` if all (compatible) overlays are saved to disk,
    ``False`` if there are any overlays with unsaved changes.
    """

    unsaved = []

    for ovl in overlayList:

        # Only Image overlays can be edited/saved
        if not isinstance(ovl, fslimage.Image): continue
        if ovl.saveState:                       continue

        unsaved.append(ovl)

    return len(unsaved) == 0



#!/usr/bin/env python
#
# glmesh.py - The GLMesh class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLMesh` class, a :class:`.GLObject` used
to render :class:`.Mesh` overlays.
"""


import numpy        as np
import numpy.linalg as npla
import OpenGL.GL    as gl

from . import                  globject
import fsl.data.utils       as dutils
import fsl.transform.affine as affine
import fsleyes.gl           as fslgl
import fsleyes.gl.routines  as glroutines
import fsleyes.gl.textures  as textures


class GLMesh(globject.GLObject):
    """The ``GLMesh`` class is a :class:`.GLObject` which encapsulates the
    logic required to draw 2D slices, and 3D renderings, of a
    :class:`.Mesh` overlay.  The ``GLMesh`` class assumes that the
    :class:`.Display` instance associated with the ``Mesh`` overlay
    holds a reference to a :class:`.MeshOpts` instance, which contains
    ``GLMesh`` specific display settings.


    **2D rendering**


    A ``GLMesh`` is rendered in one of two different ways, depending upon the
    value of the :attr:`.MeshOpts.outline` and :attr:`.MeshOpts.vertexData`
    properties - either a filled cross-secetion is drawn, or an outline of the
    cross-section is drawn.

    *Cross-sections*

    If ``outline is False and vertexData is None``, a filled cross-section of
    the mesh is drawn - see the :meth:`drawCrossSection` method.


    *Outlines*

    If ``outline is True or vertexData is not None``, an outline of the
    intersection of the mesh triangles with the viewing plane is calculated
    and drawn - see the :meth:`drawOutline` method.


    *Colouring*


    These lines will be coloured in one of the following ways:

       - With the ``MeshOpts.colour``.

       - According to the :attr:`.MeshOpts.vertexData` if it is set, in which
         case the properties of the :class:`.ColourMapOpts` class (from which
         the :class:`.MeshOpts` class derives) come into effect.

       - Or, if :attr:`vertexData` is set, and the :attr:`.MeshOpts.useLut`
         property is ``True``, the :attr:`.MeshOpts.lut` is used.


    Te ``GLMesh`` class makes use of the ``glmesh`` vertex and fragment
    shaders. These shaders are managed by two OpenGL version-specific modules -
    :mod:`.gl14.glmesh_funcs`, and :mod:`.gl21.glmesh_funcs`. These version
    specific modules must provide the following functions:

     - ``compileShaders(GLMesh)``: Compiles vertex/fragment shaders.
     - ``updateShaderState(GLMesh, **kwargs)``: Updates vertex/fragment
       shader settings.


    **3D rendering**


    3D mesh rendering is much simpler than 2D rendering. The mesh is simply
    rendered to the canvas, and coloured in the same way as described above.
    Whether the 3D mesh is coloured with a flat colour, or according to vertex
    data, shader programs are used which colour the mesh, and also apply a
    simple lighting effect.


    **Interpolation**

    When the :attr:`.MeshOpts.interpolation` property is set to ``'nearest'``,
    the mesh vertices and indices need to be modified.  In order to achieve
    nearest neighbour interpolation (a.k.a. flat shading) , each vertex of a
    triangle must be given the same vertex data value. This means that
    vertices cannot be shared by different triangles. So when the
    ``interpolation`` property changes, the :meth:`updateVertices` method will
    re-generate the vertices and indices accordingly. The
    :meth:`getVertexData` method will modify the vertex data that gets passed
    to the shader accordingly too.
    """


    def __init__(self, overlay, overlayList, displayCtx, threedee):
        """Create a ``GLMesh``.

        :arg overlay:     A :class:`.Mesh` overlay.
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
        :arg canvas:      The canvas drawing this ``GLMesh``.
        :arg threedee:    2D or 3D rendering.
        """

        globject.GLObject.__init__(
            self, overlay, overlayList, displayCtx, threedee)

        # Separate shaders are use for flat
        # colouring vs data colouring.
        self.flatShader    = None
        self.dataShader    = None
        # Two shaders are used for 2D cross
        # section drawing.
        self.xsectcpShader = None
        self.xsectblShader = None

        # We also use a render texture when
        # rendering model cross sections.
        # This texture is kept at display/
        # screen resolution
        self.renderTexture = textures.RenderTexture(
            self.name, interp=gl.GL_NEAREST)

        # Mesh overlays are coloured:
        #
        #  - with a constant colour (opts.outline == False), or
        #
        #  - with a +/- colour map, (opts.vertexData not None), or
        #
        #  - with a lookup table (opts.useLut == True and
        #    opts.vertexData is not None)
        self.cmapTexture    = textures.ColourMapTexture(  self.name)
        self.negCmapTexture = textures.ColourMapTexture(  self.name)
        self.lutTexture     = textures.LookupTableTexture(self.name)

        self.lut = None

        self.registerLut()
        self.addListeners()
        self.updateVertices()
        self.refreshCmapTextures(notify=False)

        self.compileShaders()
        self.updateShaderState()


    def destroy(self):
        """Must be called when this ``GLMesh`` is no longer needed. Removes
        some property listeners and destroys the colour map textures and
        off-screen :class:`.RenderTexture`.
        """

        self.renderTexture .destroy()
        self.cmapTexture   .destroy()
        self.negCmapTexture.destroy()
        self.lutTexture    .destroy()

        self.removeListeners()
        self.deregisterLut()

        globject.GLObject.destroy(self)

        if self.flatShader    is not None: self.flatShader   .destroy()
        if self.dataShader    is not None: self.dataShader   .destroy()
        if self.xsectcpShader is not None: self.xsectcpShader.destroy()
        if self.xsectblShader is not None: self.xsectblShader.destroy()

        self.dataShader    = None
        self.flatShader    = None
        self.xsectcpShader = None
        self.xsectblShader = None

        self.lut            = None
        self.renderTexture  = None
        self.cmapTexture    = None
        self.negCmapTexture = None
        self.lutTexture     = None


    def ready(self):
        """Overrides :meth:`.GLObject.ready`. Always returns ``True``. """
        return True


    def addListeners(self):
        """Called by :meth:`__init__`. Adds some property listeners to the
        :class:`.Display` and :class:`.MeshOpts` instances so the OpenGL
        representation can be updated when the display properties are changed..
        """

        name    = self.name
        display = self.display
        opts    = self.opts

        def shader(*a):
            self.updateShaderState()
            self.notify()

        def vertices(*a):
            self.updateVertices()
            self.updateShaderState()
            self.notify()

        def cmap(*a):
            self.refreshCmapTextures(notify=False)
            self.updateShaderState()
            self.notify()

        def lut(*a):
            self.deregisterLut()
            self.registerLut()
            self.refreshCmapTextures()

        def refresh(*a):
            self.notify()

        self.overlay.register(name, vertices, 'vertices')
        opts   .addListener('bounds',              name, vertices, weak=False)
        opts   .addListener('colour',              name, shader,   weak=False)
        opts   .addListener('outline',             name, refresh,  weak=False)
        opts   .addListener('outlineWidth',        name, refresh,  weak=False)
        opts   .addListener('wireframe',           name, refresh,  weak=False)
        opts   .addListener('vertexData',          name, shader,   weak=False)
        opts   .addListener('modulateData',        name, shader,   weak=False)
        opts   .addListener('vertexDataIndex',     name, shader,   weak=False)
        opts   .addListener('clippingRange',       name, shader,   weak=False)
        opts   .addListener('modulateRange',       name, shader,   weak=False)
        opts   .addListener('invertClipping',      name, shader,   weak=False)
        opts   .addListener('discardClipped',      name, shader,   weak=False)
        opts   .addListener('cmap',                name, cmap,     weak=False)
        opts   .addListener('useNegativeCmap',     name, cmap,     weak=False)
        opts   .addListener('negativeCmap',        name, cmap,     weak=False)
        opts   .addListener('cmapResolution',      name, cmap,     weak=False)
        opts   .addListener('interpolateCmaps',    name, cmap,     weak=False)
        opts   .addListener('invert',              name, cmap,     weak=False)
        opts   .addListener('gamma',               name, cmap,     weak=False)
        opts   .addListener('displayRange',        name, cmap,     weak=False)
        opts   .addListener('useLut',              name, shader,   weak=False)
        opts   .addListener('lut',                 name, lut,      weak=False)
        opts   .addListener('modulateAlpha',       name, shader,   weak=False)
        opts   .addListener('invertModulateAlpha', name, shader,   weak=False)
        opts   .addListener('interpolation',       name, vertices, weak=False)
        display.addListener('alpha',               name, cmap,     weak=False)

        # We don't need to listen for
        # brightness or contrast, because
        # they are linked to displayRange.


    def removeListeners(self):
        """Called by :meth:`destroy`. Removes all of the listeners added by
        the :meth:`addListeners` method.
        """
        overlay = self.overlay
        opts    = self.opts
        display = self.display
        name    = self.name

        overlay.deregister(name, 'vertices')
        opts   .removeListener('bounds',              name)
        opts   .removeListener('colour',              name)
        opts   .removeListener('outline',             name)
        opts   .removeListener('outlineWidth',        name)
        opts   .removeListener('wireframe',           name)
        opts   .removeListener('vertexData',          name)
        opts   .removeListener('modulateData',        name)
        opts   .removeListener('vertexDataIndex',     name)
        opts   .removeListener('clippingRange',       name)
        opts   .removeListener('modulateRange',       name)
        opts   .removeListener('invertClipping',      name)
        opts   .removeListener('discardClipped',      name)
        opts   .removeListener('cmap',                name)
        opts   .removeListener('useNegativeCmap',     name)
        opts   .removeListener('negativeCmap',        name)
        opts   .removeListener('cmapResolution',      name)
        opts   .removeListener('interpolateCmaps',    name)
        opts   .removeListener('invert',              name)
        opts   .removeListener('gamma',               name)
        opts   .removeListener('displayRange',        name)
        opts   .removeListener('useLut',              name)
        opts   .removeListener('lut',                 name)
        opts   .removeListener('modulateAlpha',       name)
        opts   .removeListener('invertModulateAlpha', name)
        opts   .removeListener('interpolation',       name)
        display.removeListener('alpha',               name)


    def registerLut(self):
        """Registers property listeners with the currently registered
        :class:`.LookupTable` (the :attr:`.MeshOpts.lut` property).
        """

        self.lut = self.opts.lut

        if self.lut is not None:
            for topic in ['label', 'added', 'removed']:
                self.lut.register(self.name, self.refreshCmapTextures, topic)


    def deregisterLut(self):
        """De-registers property listeners from the currently registered
        :class:`.LookupTable`.
        """

        if self.lut is not None:
            for topic in ['label', 'added', 'removed']:
                self.lut.deregister(self.name, topic)

        self.lut = None


    def updateVertices(self, *a):
        """Called by :meth:`__init__`, and when certain display properties
        change. (Re-)generates the mesh vertices, indices and normals (if
        being displayed in 3D). They are stored as attributes called
        ``vertices``, ``indices``, and ``normals`` respectively.
        """

        overlay  = self.overlay
        opts     = self.opts
        threedee = self.threedee
        vertices = overlay.vertices
        indices  = overlay.indices
        interp   = opts.interpolation
        normals  = overlay.vnormals
        vdata    = opts.getVertexData('vertex')
        xform    = opts.getTransform('mesh', 'display')

        if not np.all(np.isclose(xform, np.eye(4))):
            vertices = affine.transform(vertices, xform)

            if self.threedee:
                nmat    = affine.invert(xform).T
                normals = affine.transform(normals, nmat, vector=True)

        self.origIndices = indices
        indices          = np.asarray(indices.flatten(), dtype=np.uint32)

        # If interp == nn, we cannot share
        # vertices between triangles, so we generate
        # a set of unique vertices for each triangle,
        # and then re-generate the triangle indices.
        # The original indices are saved above, as
        # they will be used by the getVertexData
        # method to duplicate the vertex data (as
        # we need to provide a data value for each
        # vertex).
        if threedee and (vdata is not None) and interp == 'nearest':
            self.vertices = vertices[indices].astype(np.float32)
            self.indices  = np.arange(0, len(self.vertices), dtype=np.uint32)
            normals       = normals[indices, :]
        else:
            self.vertices = np.asarray(vertices, dtype=np.float32)
            self.indices  = indices

        self.vertices = dutils.makeWriteable(self.vertices)
        self.indices  = dutils.makeWriteable(self.indices)

        if self.threedee:
            self.normals = np.array(normals, dtype=np.float32)


    def frontFace(self):
        """Returns the face of the mesh triangles which which will be facing
        outwards, either ``GL_CCW`` or ``GL_CW``. This will differ depending
        on the mesh-to-display transformation matrix.
        """

        # Only looking at the mesh -> display
        # transform, thus we are assuming that
        # the MVP matrix does not have any
        # negative scales.
        xform = self.opts.getTransform('mesh', 'display')

        if npla.det(xform) > 0: return gl.GL_CCW
        else:                   return gl.GL_CW


    def getDisplayBounds(self):
        """Overrides :meth:`.GLObject.getDisplayBounds`. Returns a bounding
        box which contains the mesh vertices.
        """
        return (self.opts.bounds.getLo(),
                self.opts.bounds.getHi())


    def draw2DOutlineEnabled(self):
        """Only relevent for 2D rendering. Returns ``True`` if outline mode
        should be used, ``False`` otherwise.
        """

        opts    = self.opts
        overlay = self.overlay

        return ((overlay.trimesh is not None) and
                (opts.outline or opts.vertexData is not None))


    def preDraw(self):
        """Overrides :meth:`.GLObject.preDraw`. Binds colour map
        textures, if the mesh is being coloured with data.
        """

        if self.opts.vertexData is not None:
            if self.opts.useLut:
                self.lutTexture.bindTexture(gl.GL_TEXTURE0)
            else:
                self.cmapTexture   .bindTexture(gl.GL_TEXTURE0)
                self.negCmapTexture.bindTexture(gl.GL_TEXTURE1)


    def draw2D(self, canvas, zpos, axes, xform=None):
        """Overrids :meth:`.GLObject.draw2D`. Draws a 2D slice of the
        :class:`.Mesh`, at the specified Z location.
        """

        overlay       = self.overlay
        lo,  hi       = self.getDisplayBounds()
        xax, yax, zax = axes
        outline       = self.draw2DOutlineEnabled()

        # Mesh is 2D, and is perpendicular to
        # the viewing plane - don't draw it.
        if np.any(np.isclose([lo[xax], lo[yax]], [hi[xax], hi[yax]])):
            return

        # Mesh is 2D, and is parallel
        # to the viewing plane.
        is2D = np.isclose(lo[zax], hi[zax])

        # Don't draw this mesh if it does not
        # intersect the plane at zpos. 2D
        # meshes are always drawn, regardless
        # of the zpos.
        if not is2D and (zpos < lo[zax] or zpos > hi[zax]):
            return

        # The calculateIntersection method
        # caches cross section vertices -
        # make sure they're cleared in case
        # we are not doing an outline draw.
        self.overlayList.setData(overlay, 'crosssection_{}'.format(zax), None)

        # Delegate to the appropriate sub-method
        if   is2D:    self.draw2DMesh(canvas, xform)
        elif outline: self.drawOutline(canvas, zpos, axes, xform)
        else:         self.drawCrossSection(canvas, zpos, axes, xform)


    def draw3D(self, canvas, xform=None):
        """Overrides :meth:`.GLObject.draw3D`. Draws a 3D rendering of the
        mesh.
        """
        opts      = self.opts
        flat      = opts.vertexData is None
        mvmat     = canvas.viewMatrix
        mvpmat    = canvas.mvpMatrix
        blo, bhi  = self.getDisplayBounds()

        if flat: shader = self.flatShader
        else:    shader = self.dataShader

        if xform is not None:
            mvmat  = affine.concat(mvmat,  xform)
            mvpmat = affine.concat(mvpmat, xform)

        normmat  = affine.invert(mvmat[:3, :3]).T
        lightPos = affine.transform(canvas.lightPos, mvmat)

        is2D = np.any(np.isclose(bhi, blo))

        if opts.wireframe: polymode = gl.GL_LINE
        else:              polymode = gl.GL_FILL

        gl.glPolygonMode(gl.GL_FRONT_AND_BACK, polymode)
        gl.glLineWidth(1)

        if is2D or opts.wireframe: enable = (gl.GL_DEPTH_TEST)
        else:                      enable = (gl.GL_DEPTH_TEST, gl.GL_CULL_FACE)

        with glroutines.enabled(enable), shader.loaded():
            shader.set('MVP',       mvpmat)
            shader.set('MV',        mvmat)
            shader.set('normalmat', normmat)
            shader.set('lighting',  canvas.opts.light)
            shader.set('lightPos',  lightPos)
            gl.glFrontFace(self.frontFace())
            gl.glCullFace(gl.GL_BACK)
            shader.draw(gl.GL_TRIANGLES)

        # restore polygon mode to default (in
        # case we were drawing a wireframe)
        gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)


    def postDraw(self):
        """Overrides :meth:`.GLObject.postDraw`. """
        if self.opts.vertexData is not None:
            if self.opts.useLut:
                self.lutTexture.unbindTexture()
            else:
                self.cmapTexture   .unbindTexture()
                self.negCmapTexture.unbindTexture()


    def drawOutline(self, canvas, zpos, axes, xform=None):
        """Called by :meth:`draw2D` when ``MeshOpts.outline is True or
        MeshOpts.vertexData is not None``.  Calculates the intersection of the
        mesh with the viewing plane as a set of line segments.

        If the :attr:`.MeshOpts.outlineWidth` property is equal to ``1``, the
        segments are drawn as ``GL_LINES`` primitives. Otherwise the lines are
        converted into polygons and drawn as ``GL_TRIANGLES`` (see
        :func:`.routines.lineAsPolygon`).

        The :func:`.gl14.glmesh_funcs.draw` or
        :func:`.gl21.glmesh_funcs.draw` function called to do the draw.

        When a mesh outline is drawn on a canvas, the calculated line vertices,
        and corresponding mesh faces (triangles) are cached via the
        :meth:`.OverlayList.setData` method. This is so that other parts of
        FSLeyes can access this information if necessary (e.g. the
        :class:`.OrthoViewProfile` provides mesh-specific interaction). For a
        canvas which is showing a plane orthogonal to display axis X, Y or Z,
        this data will be given a key of ``'crosssection_0'``,
        ``'crosssection_1'``, or ``'crosssection_2'`` respectively. This cache
        is written by the :meth:`calculateIntersection` method.
        """

        opts   = self.opts
        bbox   = canvas.viewport
        flat   = opts.vertexData is None

        if flat: shader = self.flatShader
        else:    shader = self.dataShader

        # Calculate cross-section of mesh with z
        # position as a set of line segments
        vertices, faces, dists, vertXform = self.calculateIntersection(
            zpos, axes, bbox)

        if xform     is     None: xform = np.eye(4)
        if vertXform is not None: xform = affine.concat(xform, vertXform)

        xform = affine.concat(canvas.mvpMatrix, xform)

        # Retrieve vertex data associated with cros-
        # section (will return None if no vertex
        # data is selected)
        vdata    = self.getVertexData('vertex',   faces, dists)
        mdata    = self.getVertexData('modulate', faces, dists)
        vertices = vertices.reshape(-1, 3)

        # Convert line segments to triangles
        # if drawing thick lines. If default
        # line width, we can use GL_LINES
        if opts.outlineWidth == 1:
            glprim  = gl.GL_LINES
            indices = None
        else:
            zax               = axes[2]
            lineWidth         = opts.outlineWidth * canvas.pixelSize()[0]
            glprim            = gl.GL_TRIANGLES
            vertices, indices = glroutines.lineAsPolygon(
                vertices, lineWidth, zax, indices=True)

            # Each line (two vertices) is replaced with
            # two triangles (defined by four vertices).
            # So we have to duplicate each vertex data
            # value. The lineAsPolyuon function keeps
            # the vertex order the same as the original
            # line-based vertices, so we can np.repeat
            # the data.
            if vdata is not None: vdata = np.repeat(vdata, 2)
            if mdata is not None: mdata = np.repeat(mdata, 2)

        if mdata is None:
            mdata = vdata

        # Draw the outline
        with shader.loaded():
            shader.set(   'MVP',    xform)
            shader.setAtt('vertex', vertices)
            shader.setIndices(indices)
            if not flat:
                shader.setAtt('vertexData',   vdata)
                shader.setAtt('modulateData', mdata)
            shader.draw(glprim, 0, vertices.shape[0])


    def draw2DMesh(self, canvas, xform=None):
        """Not to be confused with :meth:`draw2D`.  Called by :meth:`draw2D`
        for :class:`.Mesh` overlays which are actually 2D (with a flat
        third dimension).
        """

        opts      = self.opts
        vdata     = opts.getVertexData('vertex')
        mdata     = opts.getVertexData('modulate')
        flat      = opts.vertexData is None
        vertices  = self.vertices
        indices   = self.indices

        if flat: shader = self.flatShader
        else:    shader = self.dataShader

        if xform is None: xform = canvas.mvpMatrix
        else:             xform = affine.concat(canvas.mvpMatrix, xform)

        if mdata is None:
            mdata = vdata

        # Outline mode for 2D meshes is
        # not supported at the moment.
        if opts.outline:
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
        else:
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
        with shader.loaded():
            shader.set(       'MVP',    xform)
            shader.setAtt(    'vertex', vertices)
            shader.setIndices(indices)
            if not flat:
                shader.setAtt('vertexData',   vdata)
                shader.setAtt('modulateData', mdata)
            shader.draw(gl.GL_TRIANGLES)


    def drawCrossSection(self, canvas, zpos, axes, xform=None):
        """Called by :meth:`draw2D` when ``outline is False and vertexData is None``

        Draws a filled cross-section of the mesh at the specified Z location.

        This is accomplished using a four-pass technique, roughly following
        that described at

        http://glbook.gamedev.net/GLBOOK/glbook.gamedev.net/moglgp/advclip.html

        1. The mesh back face stencil is drawn to an off-screen stencil buffer.

        2. The mesh front face stencil is subtracted from the back face
           stencil. The stencil buffer now contains a binary mask of the cross-
           section.

        3. An off-screen texture is filled with the :attr:`.MeshOpts.colour`,
           masked by the stencil buffer.

        4. The off-screen texture is drawn to the canvas.

        :arg zpos:  Position along the z axis
        :arg axes:  Tuple containing ``(x, y, z)`` axis indices.
        :arg xform: Transformation matrix to apply to the vertices.
        """

        xax, yax, zax = axes
        bbox          = canvas.viewport
        cpshader      = self.xsectcpShader
        blshader      = self.xsectblShader
        lo, hi        = self.getDisplayBounds()
        lo, hi        = self.calculateViewport(lo, hi, axes, bbox)
        xmin          = lo[xax]
        xmax          = hi[xax]
        ymin          = lo[yax]
        ymax          = hi[yax]
        tex           = self.renderTexture

        # Make sure the off-screen texture
        # matches the display size
        width, height = canvas.GetSize()
        if tex.shape != (width, height):
            tex.shape = width, height

        if xform is None: xform = canvas.mvpMatrix
        else:             xform = affine.concat(canvas.mvpMatrix, xform)

        # Figure out the equation of a plane
        # perpendicular to the Z axis, and located
        # at the z position. This is used as a
        # clipping plane to draw the mesh
        # intersection. The clip plane is defined in
        # the display coordinate system, before MVP
        # transformation.
        clipPlane                = np.zeros((4, 3), dtype=np.float32)
        clipPlane[0, [xax, yax]] = [xmin, ymin]
        clipPlane[1, [xax, yax]] = [xmin, ymax]
        clipPlane[2, [xax, yax]] = [xmax, ymax]
        clipPlane[3, [xax, yax]] = [xmax, ymin]
        clipPlane[:,  zax]       =  zpos
        clipPlane = glroutines.planeEquation(clipPlane[0, :],
                                             clipPlane[1, :],
                                             clipPlane[2, :])

        # http://glbook.gamedev.net/GLBOOK/glbook.gamedev.net/moglgp/\
        #     advclip.html
        #
        # We can calculate a cross-section of the mesh
        # by subtracting a front-face stencil from a
        # back-face stencil, clipped by the above
        # clipping plane. We use the stencil buffer of
        # the off-screen render texture, and perform
        # these steps:
        #
        # 1. Draw back faces to the stencil buffer
        # 2. Subtract front faces from the stencil buffer
        # 3. Fill the off-screen texture with the fill colour,
        #    masking it with the stencil buffer
        # 4. Draw the off-screen texture to the display
        gl.glFrontFace(self.frontFace())
        gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
        with glroutines.enabled((gl.GL_CULL_FACE, gl.GL_STENCIL_TEST)), \
             tex.target(xax, yax, lo, hi):

            glroutines.clear((0, 0, 0, 0))
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_STENCIL_BUFFER_BIT)

            # Use a clip-plane shader for steps 1 and 2
            with cpshader.loaded():

                cpshader.set('MVP',       tex.mvpMatrix)
                cpshader.set('clipPlane', clipPlane)

                # We just want to populate the stencil
                # buffer, so set the stencil test to
                # always pass, and suppress draws to
                # the rgba buffer.
                gl.glColorMask(gl.GL_FALSE, gl.GL_FALSE,
                               gl.GL_FALSE, gl.GL_FALSE)
                gl.glStencilFunc(gl.GL_ALWAYS, 0, 0)

                # first pass - draw back faces onto
                # stencil buffer. Second pass -
                # subtract front faces from back
                # faces in the stencil buffer
                faces = [gl.GL_FRONT, gl.GL_BACK]
                dirs  = [gl.GL_INCR,  gl.GL_DECR]
                for face, direction in zip(faces, dirs):
                    gl.glCullFace(face)
                    gl.glStencilOp(gl.GL_KEEP, gl.GL_KEEP, direction)
                    cpshader.draw(gl.GL_TRIANGLES)

            # Now we have a mask of the cross-section
            # in the stencil buffer - use a "blitting"
            # shader to fill the off screen texture with
            # the flat colour, but mask with the stencil
            # buffer.
            gl.glDisable(gl.GL_CULL_FACE)
            gl.glColorMask(gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE, gl.GL_TRUE)
            gl.glStencilFunc(gl.GL_NOTEQUAL, 0, 255)

            # disable blending, otherwise we will end up blending
            # with the output of the clip plane shader
            with blshader.loaded(), glroutines.disabled((gl.GL_BLEND, )):
                verts = tex.generateVertices(zpos, xmin, xmax,
                                             ymin, ymax, xax, yax)
                blshader.setAtt('vertex', verts)
                blshader.set(   'MVP',    tex.mvpMatrix)
                blshader.draw(gl.GL_TRIANGLES, 0, 6)

        # Finally, draw the off-screen texture to the display
        tex.drawOnBounds(zpos, xmin, xmax, ymin, ymax, xax, yax, xform)


    def calculateViewport(self, lo, hi, axes, bbox=None):
        """Called by :meth:`drawCrossSection`. Calculates an appropriate
        viewport (the horizontal/vertical minimums/maximums in display
        coordinates) given the ``lo`` and ``hi`` ``GLMesh`` display bounds,
        and a display ``bbox``.

        This is needed to preserve the aspect ratio of the mesh cross-section.
        """

        xax = axes[0]
        yax = axes[1]

        if bbox is not None and (lo[xax] < bbox[xax][0] or
                                 hi[xax] < bbox[xax][1] or
                                 lo[yax] > bbox[yax][0] or
                                 hi[yax] > bbox[yax][1]):

            xlen  = float(hi[xax] - lo[xax])
            ylen  = float(hi[yax] - lo[yax])
            ratio = xlen / ylen

            bblo = [ax[0] for ax in bbox]
            bbhi = [ax[1] for ax in bbox]

            lo   = [max(l, bbl) for l, bbl in zip(lo, bblo)]
            hi   = [min(h, bbh) for h, bbh in zip(hi, bbhi)]

            dxlen  = float(hi[xax] - lo[xax])
            dylen  = float(hi[yax] - lo[yax])

            dratio = dxlen / dylen

            if dratio > ratio:

                ndxlen   = xlen * (dylen / ylen)
                lo[xax] += 0.5 * (ndxlen - dxlen)
                hi[xax] -= 0.5 * (ndxlen - dxlen)

            elif dratio < ratio:

                ndylen   = ylen * (dxlen / xlen)
                lo[yax] += 0.5 * (ndylen - dylen)
                hi[yax] -= 0.5 * (ndylen - dylen)

        return lo, hi


    def calculateIntersection(self, zpos, axes, bbox=None):
        """Uses the :func:`.Mesh.planeIntersection` method to
        calculate the intersection of the mesh with the viewing plane at
        the given ``zpos``.

        :arg zpos:  Z axis coordinate at which the intersection is to be
                    calculated

        :arg axes:  Tuple containing the ``(x, y, z)`` axis indices.

        :arg bbox:  A tuple containing a ``([xlo, ylo, zlo], [xhi, yhi, zhi])``
                    bounding box to which the calculation can be restricted.

        :returns: A tuple containing:

                   - A ``(n, 2, 3)`` array which contains the two vertices of
                     a line for every intersected face (triangle) in the mesh.

                   - A ``(n, 3)`` array containing the intersected faces
                     (indices into the :attr:`.Mesh.vertices` array).

                   - A ``(n, 2, 3)`` array containing the barycentric
                     coordinates of the intersection line vertices.

                   - A ``(4, 4)`` array containing a transformation matrix
                     for transforming the line vertices into the display
                     coordinate system. May be ``None``, indicating that
                     no transformation is necessary.

        .. note:: The line vertices, and corresponding mesh triangle face
                  indices, which make up the cross section are saved
                  via the :meth:`.OverlayList.setData` method, with a key
                  ``'crosssection_[zax]'``, where ``[zax]`` is set to the
                  index of the display Z axis.
        """

        overlay     = self.overlay
        zax         = axes[2]
        opts        = self.opts
        origin      = [0] * 3
        normal      = [0] * 3
        origin[zax] = zpos
        normal[zax] = 1

        vertXform = opts.getTransform(           'mesh',    'display')
        origin    = opts.transformCoords(origin, 'display', 'mesh')
        normal    = opts.transformCoords(normal, 'display', 'mesh',
                                         vector=True)

        # TODO use bbox to constrain? This
        #      would be nice, but is not
        #      supported by trimesh.
        lines, faces, dists = overlay.planeIntersection(
            normal, origin, distances=True)

        lines = np.asarray(lines, dtype=np.float32)
        faces = np.asarray(faces, dtype=np.uint32)

        # cache the line vertices for other
        # things which might be interested.
        # See, for example, OrthoViewProfile
        # pick mode methods.
        self.overlayList.setData(overlay,
                                 'crosssection_{}'.format(zax),
                                 (lines, faces))

        faces = overlay.indices[faces]

        return lines, faces, dists, vertXform


    def getVertexData(self, vdtype, faces=None, dists=None):
        """If :attr:`.MeshOpts.vertexData` (or :attr:`.MeshOpts.modulateData`)
        is not ``None``, this method returns the vertex data to use for the
        current vertex data index.

        ``faces`` and ``dists`` are used by the :meth:`drawOutline` method,
        which draws a cross-section of the mesh. The ``dists`` array contains
        barycentric coordinates for each line vertex, and is used to linearly
        interpolate between the values of the vertices of the intersected
        triangles (defined in ``faces``).

        If ``MeshOpts.vertexData is None``, this method returns ``None``.

        :arg vdtype: Use ``'vertex'`` for :attr:`.MeshOpts.vertexData`, or
                     ``'modulate'`` for :attr:`.MeshOpts.modulateData`.
        """

        opts   = self.opts
        interp = opts.interpolation
        vdata  = opts.getVertexData(vdtype)

        if vdata is None:
            return None

        # TODO separate modulateDataIndex for modulateData?
        if vdtype == 'vertex': vdata = vdata[:, opts.vertexDataIndex]
        else:                  vdata = vdata[:, 0]

        # when using nn interp, we colour each
        # triangle according to the data value
        # for its first vertex. This should
        # really be pre-generated whenever the
        # vertex data or flat- shading option
        # is changed, but it is quick for
        # typical surfaces, so I'm not
        # bothering for now.
        if self.threedee and (vdata is not None) and interp == 'nearest':
            vdata = vdata[self.origIndices[:, 0]].repeat(3)

        # Used by drawOutline - retrieve the
        # vertex data associated with the faces
        # of the mesh that are intersected by
        # the cross section, and linearly
        # interpolate across the face if needed.
        # Otherwise colour according to the
        # first vertex in each face.
        elif faces is not None and dists is not None:
            if interp == 'linear':
                vdata = vdata[faces].repeat(2, axis=0).reshape(-1, 2, 3)
                vdata = (vdata * dists).reshape(-1, 3).sum(axis=1)
            else:
                vdata = vdata[faces[:, 0]].repeat(2)

        return np.asarray(vdata, np.float32)


    def refreshCmapTextures(self, *a, **kwa):
        """Called when various :class:`.Display` or :class:`.MeshOpts``
        properties change. Refreshes the :class:`.ColourMapTexture` instances
        corresponding to the :attr:`.MeshOpts.cmap` and
        :attr:`.MeshOpts.negativeCmap` properties, and the
        :class:`.LookupTableTexture` corresponding to the :attr:`.MeshOpts.lut`
        property.

        :arg notify: Must be passed as a keyword argument. If ``True`` (the
                     default) :meth:`.GLObject.notify` is called after the
                     textures have been updated.
        """

        notify = kwa.pop('notify', True)

        display = self.display
        opts    = self.opts
        alpha   = display.alpha / 100.0
        cmap    = opts.cmap
        interp  = opts.interpolateCmaps
        res     = opts.cmapResolution
        negCmap = opts.negativeCmap
        gamma   = opts.realGamma(opts.gamma)
        invert  = opts.invert
        dmin    = opts.displayRange[0]
        dmax    = opts.displayRange[1]

        if interp: interp = gl.GL_LINEAR
        else:      interp = gl.GL_NEAREST

        self.cmapTexture.set(cmap=cmap,
                             invert=invert,
                             alpha=alpha,
                             resolution=res,
                             gamma=gamma,
                             interp=interp,
                             displayRange=(dmin, dmax))

        self.negCmapTexture.set(cmap=negCmap,
                                invert=invert,
                                alpha=alpha,
                                resolution=res,
                                gamma=gamma,
                                interp=interp,
                                displayRange=(dmin, dmax))

        self.lutTexture.set(alpha=display.alpha           / 100.0,
                            brightness=display.brightness / 100.0,
                            contrast=display.contrast     / 100.0,
                            lut=opts.lut)

        if notify:
            self.notify()


    def compileShaders(self):
        """(Re)Compiles the vertex/fragment shader program(s), via a call
        to the GL-version specific ``compileShaders`` function.
        """
        if self.flatShader    is not None: self.flatShader   .destroy()
        if self.dataShader    is not None: self.dataShader   .destroy()
        if self.xsectcpShader is not None: self.xsectcpShader.destroy()
        if self.xsectblShader is not None: self.xsectblShader.destroy()

        fslgl.glmesh_funcs.compileShaders(self)


    def updateShaderState(self):
        """Updates the vertex/fragment shader program(s) state, via a call to
        the GL-version specific ``updateShaderState`` function.
        """

        dopts      = self.opts
        flatColour = dopts.getConstantColour()
        useNegCmap = (not dopts.useLut) and dopts.useNegativeCmap

        if dopts.useLut:
            delta     = 1.0 / (dopts.lut.max() + 1)
            cmapXform = affine.scaleOffsetXform(delta, 0.5 * delta)
        else:
            cmapXform = self.cmapTexture.getCoordinateTransform()

        modScale, modOffset = dopts.modulateScaleOffset()

        fslgl.glmesh_funcs.updateShaderState(
            self,
            useNegCmap=useNegCmap,
            cmapXform=cmapXform,
            modScale=modScale,
            modOffset=modOffset,
            flatColour=flatColour)



#!/usr/bin/env python
#
# __init__.py - FSLeyes overlay types and data-related utilities.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The :mod:`fsleyes.data` module contains FSLeyes overlay data types and
some data-related utilities.

Most FSLeyes overlay data types are defined in the ``fslpy`` library
(e.g. :class:`fsl.data.image.Image`, :class:`fsl.data.mesh.Mesh`). This
sub-package provides some additional overlay types that can be displayed with
FSLeyes:

.. autosummary::
   :nosignatures:

   ~fsleyes.data.tractogram.Tractogram
"""

import os.path as op

import fsl.utils.path as fslpath
import fsl.data.utils as dutils


def guessType(path):
    """Wrapper around the :func:`fsl.data.utils.guessType` function from the
    ``fslpy`` library, augmented to support additional data types supported
    by FSLeyes.

    :arg path: Path to a file to be loaded
    :returns:  Tuple containing:

               - A data type which can be used to load the file, or ``None``
                 if the file is not recognised.
               - The file path, possibly modified (e.g. made absolute).
    """

    import fsleyes.data.tractogram as tractogram

    dtype, path = dutils.guessType(path)

    if dtype is not None:
        return dtype, path

    if op.isfile(path):
        if fslpath.hasExt(path.lower(), tractogram.ALLOWED_EXTENSIONS):
            return tractogram.Tractogram, path

    return dtype, path



#!/usr/bin/env python
"""This package provides the :class:`.MelodicClassificationPanel`, a FSLeyes
control which allows labels to be assigned to the volumes of a 4D NIFTI image.
"""

from .melodicclassificationpanel import MelodicClassificationPanel



#!/usr/bin/env python
#
# displayspacewarning.py - The DisplaySpaceWarning class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`DisplaySpaceWarning`, a FSLeyes
control widget which displays a warning when the
:attr:`.DisplayContext.displaySpace` has a value which is preventing
the user from seeing or doing something.
"""


import                   logging
import                   wx
import wx.lib.agw.aui as wxaui

import fsl.utils.idle  as idle
import fsleyes.panel   as fslpanel
import fsleyes.strings as strings


log = logging.getLogger(__name__)


class DisplaySpaceWarning(fslpanel.FSLeyesPanel):
    """The ``DisplaySpaceWarning`` is a panel which contains a message and a
    button. When the :attr:`.DisplayContext.displaySpace` is set to a value
    that matches the ``warnCondition`` passed to ``__init__``, the warning
    message and button are shown, otherwise the entire panel is hidden. When
    the user pushes the button, the ``displaySpace`` is changed to the value
    specified by ``changeTo``.
    """

    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 frame,
                 msg,
                 warnCondition,
                 changeTo):
        """Create a ``DisplaySpaceWarning``.

        :arg parent:        ``wx`` parent object
        :arg overlayList:   The :class:`.OverlayList`
        :arg displayCtx:    The :class:`.DisplayContext`
        :arg frame:         The :class:`.FSLeyesFrame`

        :arg msg:           Message to display

        :arg warnCondition: One of ``'world'``, ``'overlay'``,
                            ``'not overlay'``, or ``'not like overlay'``

        :arg changeTo:      One of ``'world'`` or ``'overlay'``
        """

        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__warnCondition = warnCondition
        self.__changeTo      = changeTo
        self.__dsWarning     = wx.StaticText(self)
        self.__changeDS      = wx.Button(    self)

        self.__dsWarning.SetLabel(msg)
        self.__changeDS .SetLabel(strings.labels[self, 'changeDS'])

        self.__dsWarning.SetForegroundColour((192, 0, 0, 255))

        self.__sizer     = wx.BoxSizer(wx.HORIZONTAL)
        self.__realSizer = wx.BoxSizer(wx.HORIZONTAL)

        self.__realSizer.Add((1, 1),           flag=wx.EXPAND)
        self.__realSizer.Add(self.__dsWarning)
        self.__realSizer.Add((10, 1))
        self.__realSizer.Add(self.__changeDS,  flag=wx.ALIGN_CENTRE_VERTICAL)
        self.__realSizer.Add((1, 1),           flag=wx.EXPAND)
        self.__sizer    .Add(self.__realSizer, flag=wx.EXPAND)

        self.SetSizer(self.__sizer)
        self.Fit()
        self.__size = self.GetBestSize()

        self.__changeDS .Bind(wx.EVT_BUTTON, self.__onChangeDS)

        displayCtx .addListener('displaySpace',
                                self.name,
                                self.__displaySpaceChanged)
        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__displaySpaceChanged)
        overlayList.addListener('overlays',
                                self.name,
                                self.__displaySpaceChanged)

        idle.idle(self.__displaySpaceChanged)


    def destroy(self):
        """Must be called when this ``DisplaySpaceWarning`` is no longer
        needed. De-registers listeners.
        """

        displayCtx  = self.displayCtx
        overlayList = self.overlayList

        displayCtx .removeListener('displaySpace',    self.name)
        displayCtx .removeListener('selectedOverlay', self.name)
        overlayList.removeListener('overlays',        self.name)

        fslpanel.FSLeyesPanel.destroy(self)


    def __displaySpaceChanged(self, *a):
        """Called when the :attr:`.DisplayContext.displaySpace` property
        changes. If it has been given a setting that would cause the
        transformation changes to have no effect on the display, a warning
        message is shown.
        """

        if self.destroyed:
            return

        parent       = self.GetParent()
        condition    = self.__warnCondition
        displayCtx   = self.displayCtx
        displaySpace = displayCtx.displaySpace
        overlay      = displayCtx.getSelectedOverlay()

        if overlay is not None:
            overlay = displayCtx.getOpts(overlay).referenceImage

        if   condition == 'overlay':     show = displaySpace is overlay
        elif condition == 'not overlay': show = displaySpace is not overlay
        elif condition == 'world':       show = displaySpace == 'world'
        elif condition == 'like overlay':
            if displaySpace == 'world':
                show = False
            else:
                show = (overlay is not None) and \
                    overlay.sameSpace(displaySpace)
        elif condition == 'not like overlay':
            if displaySpace == 'world':
                show = True
            else:
                show = (overlay is not None) and \
                    (not overlay.sameSpace(displaySpace))
        else:
            show = False

        if show: size = self.__size
        else:    size = (0, 0)

        if show:
            log.debug('Showing display space warning ({} / {})'.format(
                displaySpace, condition))

        self.SetMinSize(   size)
        self.CacheBestSize(size)

        self.__sizer.Show(self.__realSizer, show)
        self.Layout()
        self.Fit()

        sizer = parent.GetSizer()
        parent.SetInitialSize(sizer.GetMinSize())

        parent = parent.GetTopLevelParent()
        sizer  = parent.GetSizer()
        if isinstance(parent, wxaui.AuiFloatingFrame):
            parent.SetInitialSize(sizer.GetMinSize())
            parent.Layout()
            parent.Fit()


    def __onChangeDS(self, ev):
        """Called when the *Change display space* button is pushed. This button
        is only shown if the :attr:`.DisplayContext.displaySpace` is set to
        something which causes the transformation change to have no effect on
        the display. This method changes the ``displaySpace`` to that
        specified by the ``changeTo`` argument passed to :meth:`__init__`.
        """

        changeTo   = self.__changeTo
        displayCtx = self.displayCtx

        if changeTo == 'world':
            newSpace = 'world'
        elif changeTo == 'overlay' and len(self.overlayList) > 0:
            newSpace = displayCtx.getSelectedOverlay()
            newSpace = displayCtx.getReferenceImage(newSpace)

        log.debug('Changing displaySpace to {}'.format(newSpace))
        displayCtx.displaySpace = newSpace



#!/usr/bin/env python
#
# glmip.py -  Maximum intensity projection
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The :class:`GLMIP` class can be used to render
maximum-intensity-projections of an :class:`.Image` overlay onto a 2D canvas.
"""


import OpenGL.GL                 as gl

import fsl.utils.idle            as idle

import fsleyes.gl                as fslgl
import fsleyes.gl.textures       as textures
import fsleyes.gl.resources      as glresources
from . import                       glimageobject


class GLMIP(glimageobject.GLImageObject):
    """The ``GLMIP`` class is a :class:`.GLImageObject` which can be used to
    render maximum-intensity-projections of an :class:`.Image` overlay onto a
    2D canvas.

    There is no support for rending a MIP onto a 3D canvas, as the
    :class:`.GLVolume` can be used to achieve a MIP-like effect.

    To use the ``GLMIP``, the :class:`.Display.overlayType` attribute for the
    image must be set to ``'mip'``. See the :class:`.MIPOpts` class for more
    details.

    The ``GLMIP`` class uses functions defined in the :mod:`.gl21.glmip_funcs`
    module - there is currently no support for OpenGL 1.4.
    """


    def __init__(self, image, overlayList, displayCtx, threedee):
        """Create a ``GLMIP``.

        :arg image:       An :class:`.Image` object.

        :arg overlayList: The :class:`.OverlayList`

        :arg displayCtx:  The :class:`.DisplayContext` object managing the
                          scene.

        :arg threedee:    Set up for 2D or 3D rendering.
        """

        glimageobject.GLImageObject.__init__(self,
                                             image,
                                             overlayList,
                                             displayCtx,
                                             threedee)

        self.shader         = None
        self.imageTexture   = None
        self.cmapTexture    = textures.ColourMapTexture(self.name)

        self.addDisplayListeners()
        self.refreshImageTexture()
        self.refreshCmapTextures()

        def init():
            fslgl.glmip_funcs.init(self)
            self.notify()

        idle.idleWhen(init, self.textureReady)


    def destroy(self):
        """Clears up resources used by the ``GLMIP``. """

        self.cmapTexture.destroy()

        self.removeDisplayListeners()
        self.imageTexture.deregister(self.name)
        glresources.delete(self.imageTexture.name)

        fslgl.glmip_funcs.destroy(self)
        glimageobject.GLImageObject.destroy(self)

        self.shader         = None
        self.cmapTexture    = None
        self.imageTexture   = None


    def addDisplayListeners(self):
        """Adds a bunch of listeners to the :class:`.Display` object, and the
        associated :class:`.MIPOpts` instance, which define how the image
        should be displayed.
        """
        display = self.display
        opts    = self.opts
        name    = self.name

        def shader(*a):
            self.updateShaderState()

        def cmap(*a):
            self.refreshCmapTextures()
            self.updateShaderState(alwaysNotify=True)

        opts    .addListener('window',           name, shader,  weak=False)
        opts    .addListener('minimum',          name, shader,  weak=False)
        opts    .addListener('absolute',         name, shader,  weak=False)
        opts    .addListener('displayRange',     name, cmap,    weak=False)
        opts    .addListener('clippingRange',    name, shader,  weak=False)
        opts    .addListener('invertClipping',   name, shader,  weak=False)
        display .addListener('alpha',            name, cmap,    weak=False)
        opts    .addListener('cmap',             name, cmap,    weak=False)
        opts    .addListener('gamma',            name, cmap,    weak=False)
        opts    .addListener('interpolateCmaps', name, cmap,    weak=False)
        opts    .addListener('negativeCmap',     name, cmap,    weak=False)
        opts    .addListener('cmapResolution',   name, cmap,    weak=False)
        opts    .addListener('useNegativeCmap',  name, cmap,    weak=False)
        opts    .addListener('invert',           name, cmap,    weak=False)
        opts    .addListener('volume',           name, self.__volumeChanged)
        opts    .addListener('interpolation',    name, self.__interpChanged)
        opts    .addListener('transform',        name, self.notify)
        opts    .addListener('displayXform',     name, self.notify)

        # See comment in GLVolume.addDisplayListeners about this
        self.__syncListenersRegistered = opts.getParent() is not None

        if self.__syncListenersRegistered:
            opts.addSyncChangeListener(
                'volume', name, self.refreshImageTexture)


    def removeDisplayListeners(self):
        """Removes all listeners added by :meth:`addDisplayListeners`. """

        display = self.display
        opts    = self.opts
        name    = self.name

        if self.__syncListenersRegistered:
            opts.removeSyncChangeListener('volume', name)

        opts    .removeListener('window',           name)
        opts    .removeListener('minimum',          name)
        opts    .removeListener('absolute',         name)
        opts    .removeListener('displayRange',     name)
        opts    .removeListener('clippingRange',    name)
        opts    .removeListener('invertClipping',   name)
        display .removeListener('alpha',            name)
        opts    .removeListener('cmap',             name)
        opts    .removeListener('gamma',            name)
        opts    .removeListener('interpolateCmaps', name)
        opts    .removeListener('negativeCmap',     name)
        opts    .removeListener('cmapResolution',   name)
        opts    .removeListener('useNegativeCmap',  name)
        opts    .removeListener('invert',           name)
        opts    .removeListener('volume',           name)
        opts    .removeListener('interpolation',    name)
        opts    .removeListener('transform',        name)
        opts    .removeListener('displayXform',     name)


    def refreshImageTexture(self):
        """Makes sure that the :class:`.ImageTexture`, used to store the
        :class:`.Image` data, is up to date.
        """

        opts     = self.opts
        texName  = '{}_{}' .format(type(self).__name__, id(self.image))
        unsynced = (opts.getParent() is None or
                    not opts.isSyncedToParent('volume'))

        if unsynced:
            texName = '{}_unsync_{}'.format(texName, id(opts))

        if self.imageTexture is not None:
            if self.imageTexture.name == texName:
                return

            self.imageTexture.deregister(self.name)
            glresources.delete(self.imageTexture.name)

        if opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                            interp = gl.GL_LINEAR

        self.imageTexture = glresources.get(
            texName,
            textures.ImageTexture,
            texName,
            self.image,
            interp=interp,
            volume=opts.index()[3:],
            notify=False)

        self.imageTexture.register(self.name, self.__imageTextureChanged)


    def refreshCmapTextures(self):
        """Updates the colour map texture in line with the current
        :class:`.Display` and :class:`MIPOpts` settings.
        """
        display = self.display
        opts    = self.opts
        alpha   = display.alpha / 100.0
        cmap    = opts.cmap
        interp  = opts.interpolateCmaps
        res     = opts.cmapResolution
        gamma   = opts.realGamma(opts.gamma)
        invert  = opts.invert
        dmin    = opts.displayRange[0]
        dmax    = opts.displayRange[1]

        if interp: interp = gl.GL_LINEAR
        else:      interp = gl.GL_NEAREST

        self.cmapTexture.set(cmap=cmap,
                             invert=invert,
                             alpha=alpha,
                             resolution=res,
                             gamma=gamma,
                             interp=interp,
                             displayRange=(dmin, dmax))


    def updateShaderState(self, *args, **kwargs):
        """Calls :func:`.gl21.glmip_funcs.updateShaderState`, and
        :meth:`.Notifier.notify`. Uses :func:`.idle.idleWhen` to ensure that
        they don't get called until :meth:`ready` returns ``True``.
        """
        alwaysNotify = kwargs.pop('alwaysNotify', None)

        def func():
            if fslgl.glmip_funcs.updateShaderState(self) or alwaysNotify:
                self.notify()

        idle.idleWhen(func,
                      self.ready,
                      name=self.name,
                      skipIfQueued=True)


    def ready(self):
        """Returns ``True`` if this ``GLMIP`` is ready to be drawn, ``False``
        otherwise.
        """
        return self.shader is not None and self.textureReady()


    def textureReady(self):
        """Returns ``True`` if the ``imageTexture`` is ready to be used,
        ``False`` otherwise.
        """
        return self.imageTexture is not None and self.imageTexture.ready()


    def preDraw(self):
        """Binds textures. """
        self.imageTexture.bindTexture(gl.GL_TEXTURE0)
        self.cmapTexture .bindTexture(gl.GL_TEXTURE1)


    def draw2D(self, *args, **kwargs):
        """Calls :func:`.gl21.glmip_funcs.draw2D`. """
        fslgl.glmip_funcs.draw2D(self, *args, **kwargs)


    def drawAll(self, *args, **kwargs):
        """Calls :func:`.gl21.glmip_funcs.draw2D`. """
        fslgl.glmip_funcs.drawAll(self, *args, **kwargs)


    def draw3D(self, *args, **kwargs):
        """Does nothing - not supported. """


    def postDraw(self):
        """Unbinds textures. """
        self.imageTexture  .unbindTexture()
        self.cmapTexture   .unbindTexture()


    def __volumeChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.volume` property changes. Updates
        the image texture accordingly.
        """
        self.imageTexture.set(volume=self.opts.index()[3:])


    def __interpChanged(self, *a):
        """Called when the :attr:`.MIPOpts.interpolation` changes. Updates the
        image texture.
        """
        if self.opts.interpolation == 'none': interp = gl.GL_NEAREST
        else:                                 interp = gl.GL_LINEAR
        self.imageTexture.set(interp=interp)


    def __imageTextureChanged(self, *a):
        """Called when the image texture data has changed. Triggers a refresh.
        """
        self.updateShaderState(alwaysNotify=True)


    def __imageSyncChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.volume` property is synchronised
        or un-synchronised. Calls :meth:`refreshImageTexture` and
        :meth:`updateShaderState`.
        """
        self.refreshImageTexture()
        self.updateShaderState(alwaysNotify=True)



#!/usr/bin/env python
#
# rendertexture.py - The RenderTexture and GLObjectRenderTexture classes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`RenderTexture` and
:class:`GLObjectRenderTexture` classes, which are :class:`.Texture2D`
sub-classes intended to be used as targets for off-screen rendering.

These classes are used by the :class:`.SliceCanvas` and
:class:`.LightBoxCanvas` classes for off-screen rendering, and in various
other situations throughout FSLeyes.
"""


import logging
import contextlib

import numpy                             as np
import OpenGL.GL                         as gl
import OpenGL.raw.GL._types              as gltypes

import fsl.transform.affine   as affine
import fsleyes.gl             as fslgl
import fsleyes.gl.extensions  as glexts
import fsleyes.gl.routines    as glroutines
import fsleyes.gl.shaders     as shaders
from . import                    texture2d


log = logging.getLogger(__name__)


class RenderTexture(texture2d.Texture2D):
    """The ``RenderTexture`` class is a 2D RGBA texture which manages a frame
    buffer, and a render buffer or a :class:`.DepthTexture` which is used as
    the depth attachment.

    A ``RenderTexture`` is intended to be used as a target for off-screen
    rendering. Using a ``RenderTexture`` (``tex`` in the example below) as the
    rendering target is easy::

        # Set the texture size in pixels
        tex.shape = 1024, 768

        # Bind the texture/frame buffer, and configure
        # the viewport for orthoghraphic display.
        lo = (0.0, 0.0, 0.0)
        hi = (1.0, 1.0, 1.0)

        with tex.target(0, 1, lo, hi):
            # ...
            # draw the scene
            # ...


    The contents of the ``RenderTexture`` can later be drawn to the screen
    via the :meth:`.Texture2D.draw` or :meth:`.Texture2D.drawOnBounds`
    methods.


    A ``RenderTexture`` can be configured in one of three ways:

      1. Using a ``RGBA8`` :class:`.Texture2D` (the ``RenderTexture`` itself)
         as the colour attachment, and no depth or stencil attachments.

      2. As above for the colour attachment, and a :class:`.DepthTexture` as
         the depth attachment.

      3. As above for the colour attachment, and a ``DEPTH24_STENCIL8``
         renderbuffer as the combined depth+stencil attachment.


    These are the only available options due to limitations in different GL
    implementations. You can choose which option you wish to use via the
    ``rttype`` argument to ``__init__``.

    If you choose option #2, a :class:`.DepthTexture` instance will be used as
    the depth attachment. The resulting depth values created after drawing to
    the ``RenderTexture`` may be used in subsequent processing.  You can
    either access the depth texture directly via the :meth:`depthTexture`
    method, or you can draw the contents of the ``RenderTexture``, with depth
    information, by passing the ``useDepth`` flag to either the :meth:`draw`
    or :meth:`drawOnBounds` methods.
    """


    def __init__(self, name, rttype='cds', **kwargs):
        """Create a ``RenderTexture``. All keyword arguments are passed through
        to the :meth:`.Texture2D.__init__` method.

        :arg name:   Unique name for this texture

        :arg rttype: ``RenderTexture`` configuration. If provided, must be
                     passed as a keyword argument. Valid values are:

                       - ``'c'``:    Only a colour attachment is configured
                       - ``'cd'``:   Colour and depth attachments are
                                     configured
                       - ``'cds'`` : Colour, depth, and stencil attachments
                                     are configured. This is the default.

        .. note:: A rendering target must have been set for the GL context
                  before a frame buffer can be created ... in other words,
                  call ``context.SetCurrent`` before creating a
                  ``RenderTexture``.
        """

        if rttype not in ('c', 'cd', 'cds'):
            raise ValueError('Invalid rttype: {}'.format(rttype))

        texture2d.Texture2D.__init__(self,
                                     name,
                                     ndim=2,
                                     nvals=4,
                                     dtype=np.uint8,
                                     **kwargs)

        self.__frameBuffer      = glexts.glGenFramebuffers(1)
        self.__rttype           = rttype
        self.__renderBuffer     = None
        self.__depthTexture     = None
        self.__oldFrameBuffer   = None
        self.__oldRenderBuffer  = None
        self.__oldSize          = None
        self.__projectionMatrix = None
        self.__viewMatrix       = None
        self.__viewport         = None

        # Use a single renderbuffer as
        # the depth+stencil attachment
        if rttype == 'cds':
            self.__renderBuffer = glexts.glGenRenderbuffers(1)

        # Or use a texture as the depth
        # attachment (with no stencil attachment)
        elif rttype == 'cd':
            self.__depthTexture = texture2d.DepthTexture(
                '{}_depth'.format(self.name))

            # We also need a shader program in
            # case the creator is intending to
            # use the depth information in draws.
            self.__initShader()

        log.debug('Created fbo %s [%s]', self.__frameBuffer, rttype)


    def __initShader(self):
        """Called by :meth:`__init__` if this ``RenderTexture`` was
        configured to use a colour and depth texture. Compiles
        vertex/fragment shader programs which pass the colour and depth
        values through.

        These shaders are used if the :meth:`draw` or
        :meth:`.Texture2D.drawOnBounds` methods are used with the
        ``useDepth=True`` argument.
        """

        self.__shader = None

        vertSrc   = shaders.getVertexShader(  'rendertexture')
        fragSrc   = shaders.getFragmentShader('rendertexture')

        if float(fslgl.GL_COMPATIBILITY) < 2.1:
            self.__shader = shaders.ARBPShader(vertSrc,
                                               fragSrc,
                                               {'colourTexture' : 0,
                                                'depthTexture'  : 1})
        else:
            self.__shader = shaders.GLSLShader(vertSrc, fragSrc)
            self.__shader.load()
            self.__shader.set('colourTexture', 0)
            self.__shader.set('depthTexture',  1)
            self.__shader.unload()


    def destroy(self):
        """Must be called when this ``RenderTexture`` is no longer needed.
        Destroys the frame buffer and render buffer, and calls
        :meth:`.Texture2D.destroy`.
        """

        texture2d.Texture2D.destroy(self)

        log.debug('Deleting FBO%s [%s]', self.__frameBuffer, self.__rttype)

        glexts.glDeleteFramebuffers(1, gltypes.GLuint(self.__frameBuffer))

        if self.__renderBuffer is not None:
            rb = gltypes.GLuint(self.__renderBuffer)
            glexts.glDeleteRenderbuffers(1, rb)

        if self.__depthTexture is not None:
            self.__depthTexture.destroy()

        self.__frameBuffer     = None
        self.__renderBuffer    = None
        self.__depthTexture    = None
        self.__oldFrameBuffer  = None
        self.__oldRenderBuffer = None


    @property
    def projectionMatrix(self):
        """Return the projection matrix to use when drawing to this
        ``RenderTexture``. Only returns a value when this ``RenderTexture``
        is bound - returns ``None`` at all other times.
        """
        return self.__projectionMatrix


    @property
    def viewMatrix(self):
        """Return the model-view matrix to use when drawing to this
        ``RenderTexture``. Only returns a value when this ``RenderTexture``
        is bound - returns ``None`` at all other times.
        """
        return self.__viewMatrix


    @property
    def mvpMatrix(self):
        """Returns the current model*view*projection matrix. """
        return affine.concat(self.__projectionMatrix, self.__viewMatrix)


    @property
    def viewport(self):
        """Return the display coordinate system bounding box for this
        ``RenderTexture`` as a sequence of three ``(low, high)`` tuples.
        Only returns a value when this ``RenderTexture`` is bound - returns
        ``None`` at all other times.
        """
        return self.__viewport


    @texture2d.Texture2D.data.setter
    def data(self, data):
        """Raises a :exc:`NotImplementedError`. The ``RenderTexture`` derives
        from the :class:`.Texture2D` class, but is not intended to have its
        texture data manually set - see the :class:`.Texture2D` documentation.
        """
        raise NotImplementedError('Texture data cannot be set for {} '
                                  'instances'.format(type(self).__name__))


    @property
    def depthTexture(self):
        """Returns the ``DepthTexture`` instance used as the depth buffer.
        Returns ``None`` if this ``RenderTexture`` was not configured with
        ``'cd'`` (see :meth:`__init__`).
        """
        return self.__depthTexture


    @depthTexture.setter
    def depthTexture(self, dtex):
        """Replace the depth texture currently used by this ``RenderTexture``.

        A :exc:`ValueError` is raised if ``dtex`` is not compatible with this
        ``RenderTexture``.

        :arg dtex: A :class:`.DepthTexture` instance
        """

        if self.__depthTexture is None:
            raise ValueError('This RenderTexture is not '
                             'configured to use a depth texture')

        if not isinstance(dtex, texture2d.DepthTexture)    or \
           dtex.internalFormat != gl.GL_DEPTH_COMPONENT24  or \
           dtex.shape          != self.shape:
            raise ValueError('Incompatible depth texture')

        self.__depthTexture = dtex


    @texture2d.Texture2D.shape.setter
    def shape(self, shape):
        """Overrides the :meth:`.Texture2D.shape` setter. Calls that method, and
        also calls it on the depth texture if it exists.
        """

        width, height = shape

        # We have to size the depth texture first,
        # because calling shape on ourselves will
        # result in refresh being called, which
        # expects the depth texture to be ready to
        # go.
        if self.__depthTexture is not None:
            self.__depthTexture.shape = width, height

        texture2d.Texture2D.shape.fset(self, (width, height))


    @contextlib.contextmanager
    def renderViewport(self, xax, yax, lo, hi):
        """Context manager which sets and restores the viewport via
        :meth:`setRenderViewport` and :meth:`restoreViewport`.
        """

        self.setRenderViewport(xax, yax, lo, hi)
        try:
            yield
        finally:
            self.restoreViewport()


    def setRenderViewport(self, xax, yax, lo, hi):
        """Configures the GL viewport for a 2D orthographic display. See the
        :func:`.routines.show2D` function.

        The existing viewport settings are cached, and can be restored via
        the :meth:`restoreViewport` method.

        :arg xax: The display coordinate system axis which corresponds to the
                  horizontal screen axis.

        :arg yax: The display coordinate system axis which corresponds to the
                  vertical screen axis.

        :arg lo:  A tuple containing the minimum ``(x, y, z)`` display
                  coordinates.

        :arg hi:  A tuple containing the maximum ``(x, y, z)`` display
                  coordinates.
        """

        if self.__oldSize is not None:
            raise RuntimeError('RenderTexture RB{}/FBO{} has already '
                               'configured the viewport'.format(
                                   self.__renderBuffer,
                                   self.__frameBuffer))

        log.debug('Configuring viewport for RB%s/FBO%s',
                  self.__renderBuffer, self.__frameBuffer)

        self.__oldSize = gl.glGetIntegerv(gl.GL_VIEWPORT)

        width, height  = self.shape
        projmat, mvmat = glroutines.show2D(xax, yax, lo, hi)

        gl.glViewport(0, 0, width, height)

        self.__viewport         = list(zip(lo, hi))
        self.__projectionMatrix = projmat
        self.__viewMatrix       = mvmat


    def restoreViewport(self):
        """Restores the GL viewport settings which were saved via a prior call
        to :meth:`setRenderViewport`.
        """

        if self.__oldSize is None:
            raise RuntimeError('RenderTexture RB{}/FBO{} has not '
                               'configured the viewport'.format(
                                   self.__renderBuffer,
                                   self.__frameBuffer))

        log.debug('Clearing viewport (from RB%s/FBO%s)',
                  self.__renderBuffer, self.__frameBuffer)

        gl.glViewport(*self.__oldSize)
        self.__oldSize          = None
        self.__projectionMatrix = None
        self.__viewMatrix       = None
        self.__viewport         = None


    @contextlib.contextmanager
    def target(self, *args, **kwargs):
        """Context manager which binds and unbinds this ``RenderTexture`` as
        the render target, via :meth:`bindAsRenderTarget` and
        :meth:`unbindAsRenderTarget`.

        If any arguments are provided, the viewport is also set and restored
        via :meth:`setRenderViewport` and :meth:`restoreViewport`.
        """

        setViewport  = len(args) > 0 or len(kwargs) > 0
        alreadyBound = self.__oldFrameBuffer is not None

        if not alreadyBound:
            self.bindAsRenderTarget()

        if setViewport:
            self.setRenderViewport(*args, **kwargs)
        try:
            yield
        finally:
            if setViewport:
                self.restoreViewport()
            if not alreadyBound:
                self.unbindAsRenderTarget()


    def bindAsRenderTarget(self):
        """Configures the frame buffer and render buffer of this
        ``RenderTexture`` as the targets for rendering.

        The existing farme buffer and render buffer are cached, and can be
        restored via the :meth:`unbindAsRenderTarget` method.
        """

        if self.__oldFrameBuffer is not None:
            raise RuntimeError('RenderTexture FBO{} is already '
                               'bound'.format(self.__frameBuffer))

        self.__oldFrameBuffer  = gl.glGetIntegerv(
            glexts.GL_FRAMEBUFFER_BINDING)

        if self.__renderBuffer is not None:
            self.__oldRenderBuffer = gl.glGetIntegerv(
                glexts.GL_RENDERBUFFER_BINDING)

        log.debug('Setting FBO%s as render target', self.__frameBuffer)

        glexts.glBindFramebuffer(
            glexts.GL_FRAMEBUFFER, self.__frameBuffer)

        if self.__renderBuffer is not None:
            glexts.glBindRenderbuffer(
                glexts.GL_RENDERBUFFER, self.__renderBuffer)


    def unbindAsRenderTarget(self):
        """Restores the frame buffer and render buffer which were saved via a
        prior call to :meth:`bindAsRenderTarget`.
        """

        if self.__oldFrameBuffer is None:
            raise RuntimeError('RenderTexture FBO{} has not been '
                               'bound'.format(self.__frameBuffer))

        log.debug('Restoring render target to FBO%s (from FBO%s)',
                  self.__oldFrameBuffer, self.__frameBuffer)

        glexts.glBindFramebuffer(
            glexts.GL_FRAMEBUFFER, self.__oldFrameBuffer)

        if self.__renderBuffer is not None:
            glexts.glBindRenderbuffer(
                glexts.GL_RENDERBUFFER, self.__oldRenderBuffer)

        self.__oldFrameBuffer  = None
        self.__oldRenderBuffer = None


    def doRefresh(self):
        """Overrides :meth:`.Texture2D.doRefresh`. Calls the base-class
        implementation, and ensures that the frame buffer and render buffer
        of this ``RenderTexture`` are configured correctly.
        """
        texture2d.Texture2D.doRefresh(self)

        width, height = self.shape

        log.debug('Refreshing render texture FBO%s', self.__frameBuffer)

        # Bind the colour buffer
        with self.target():
            glexts.glFramebufferTexture2D(
                glexts.GL_FRAMEBUFFER,
                glexts.GL_COLOR_ATTACHMENT0,
                gl   .GL_TEXTURE_2D,
                self.handle,
                0)

            # Combined depth/stencil attachment
            if self.__rttype == 'cds':

                # Configure the render buffer
                glexts.glRenderbufferStorage(
                    glexts.GL_RENDERBUFFER,
                    gl.GL_DEPTH24_STENCIL8,
                    width,
                    height)

                # Bind the render buffer
                glexts.glFramebufferRenderbuffer(
                    glexts.GL_FRAMEBUFFER,
                    gl.GL_DEPTH_STENCIL_ATTACHMENT,
                    glexts.GL_RENDERBUFFER,
                    self.__renderBuffer)

            # Or a depth texture
            elif self.__rttype == 'cd':
                glexts.glFramebufferTexture2D(
                    glexts.GL_FRAMEBUFFER,
                    glexts.GL_DEPTH_ATTACHMENT,
                    gl   .GL_TEXTURE_2D,
                    self.__depthTexture.handle,
                    0)

            # Get the FBO status before unbinding it -
            # the Apple software renderer will return
            # FRAMEBUFFER_UNDEFINED otherwise.
            status = glexts.glCheckFramebufferStatus(
                glexts.GL_FRAMEBUFFER)

        # Complain if something is not right
        if status != glexts.GL_FRAMEBUFFER_COMPLETE:
            raise RuntimeError('An error has occurred while configuring '
                               'the frame buffer [{}]'.format(status))


    def draw(self, vertices, useDepth=False):
        """Overrides :meth:`.Texture2D.draw`. Calls that method, optionally
        using the information in the depth texture.


        :arg useDepth: Defaults to ``False``. If ``True``, and this
                       ``RenderTexture`` was configured to use a depth
                       texture, the texture is rendered with depth
                       information using a fragment program.


        A ``RuntimeError`` will be raised if ``useDepth is True``, but this
        ``RenderTexture`` was not configured appropriately (the ``'cd'``
        setting in :meth:`__init__`).
        """

        if useDepth and self.__depthTexture is None:
            raise RuntimeError('useDepth is True but I don\'t '
                               'have a depth texture!')

        # Draw using the default Texture2D shader
        if not useDepth:
            texture2d.Texture2D.draw(self, vertices)
        else:
            shader    = self.__shader
            depthTex  = self.__depthTexture
            texCoords = self.generateTextureCoords()
            nverts    = vertices.shape[0]
            with shader.loaded():
                shader.setAtt('vertex',   vertices)
                shader.setAtt('texCoord', texCoords)
                with self    .bound(gl.GL_TEXTURE0), \
                     depthTex.bound(gl.GL_TEXTURE1):
                    shader.draw(gl.GL_TRIANGLES, 0, nverts)


class GLObjectRenderTexture(RenderTexture):
    """The ``GLObjectRenderTexture`` is a :class:`RenderTexture` intended to
    be used for rendering :class:`.GLObject` instances off-screen.


    The advantage of using a ``GLObjectRenderTexture`` over a
    :class:`.RenderTexture` is that a ``GLObjectRenderTexture`` will
    automatically adjust its size to suit the resolution of the
    :class:`.GLObject` - see the :meth:`.GLObject.getDataResolution` method.


    In order to accomplish this, the :meth:`setAxes` method must be called
    whenever the display orientation changes, so that the render texture
    size can be re-calculated.
    """

    def __init__(self, name, globj, xax, yax, maxResolution=2048):
        """Create a ``GLObjectRenderTexture``.

        :arg name:          A unique name for this ``GLObjectRenderTexture``.

        :arg globj:         The :class:`.GLObject` instance which is to be
                            rendered.

        :arg xax:           Index of the display coordinate system axis to be
                            used as the horizontal render texture axis.

        :arg yax:           Index of the display coordinate system axis to be
                            used as the vertical render texture axis.

        :arg maxResolution: Maximum resolution in pixels, along either the
                            horizontal or vertical axis, for this
                            ``GLObjectRenderTexture``.
        """

        self.__globj         = globj
        self.__xax           = xax
        self.__yax           = yax
        self.__maxResolution = maxResolution

        RenderTexture.__init__(self, name)

        name = '{}_{}'.format(self.name, id(self))
        globj.register(name, self.__updateShape)

        self.__updateShape()


    def destroy(self):
        """Must be called when this ``GLObjectRenderTexture`` is no longer
        needed. Removes the update listener from the :class:`.GLObject`, and
        calls :meth:`.RenderTexture.destroy`.
        """

        name = '{}_{}'.format(self.name, id(self))
        self.__globj.deregister(name)
        RenderTexture.destroy(self)


    def setAxes(self, xax, yax):
        """This method must be called when the display orientation of the
        :class:`GLObject` changes. It updates the size of this
        ``GLObjectRenderTexture`` so that the resolution and aspect ratio
        of the ``GLOBject`` are maintained.
        """
        self.__xax = xax
        self.__yax = yax
        self.__updateShape()


    @RenderTexture.shape.setter
    def shape(self, shape):
        """Overrides the :meth:`.Texture.shape` setter. Raises a
        :exc:`NotImplementedError`. The size of a ``GLObjectRenderTexture`` is
        set automatically.
        """
        raise NotImplementedError(
            'Texture size cannot be set for {} instances'.format(
                type(self).__name__))


    def __updateShape(self, *a):
        """Updates the size of this ``GLObjectRenderTexture``, basing it
        on the resolution returned by the :meth:`.GLObject.getDataResolution`
        method. If that method returns ``None``, a default resolution is used.
        """
        globj  = self.__globj
        maxRes = self.__maxResolution

        resolution = globj.getDataResolution(self.__xax, self.__yax)

        # Default resolution is based on the canvas size
        if resolution is None:

            size                   = gl.glGetIntegerv(gl.GL_VIEWPORT)
            width                  = size[2]
            height                 = size[3]
            resolution             = [100] * 3
            resolution[self.__xax] = width
            resolution[self.__yax] = height

            log.debug('Using default resolution for GLObject %s: %s',
                      type(globj).__name__, resolution)

        width  = resolution[self.__xax]
        height = resolution[self.__yax]

        if any((width <= 0, height <= 0)):
            raise ValueError('Invalid GLObject resolution: {}'.format(
                (width, height)))

        if width > maxRes or height > maxRes:
            ratio = min(width, height) / float(max(width, height))

            if width > height:
                width  = maxRes
                height = width * ratio
            else:
                height = maxRes
                width  = height * ratio

            width  = int(round(width))
            height = int(round(height))

        log.debug('Setting %s texture resolution to %sx%s',
                  type(globj).__name__, width, height)

        RenderTexture.shape.fset(self, (width, height))



#!/usr/bin/env python
#
# glmesh_funcs.py - OpenGL 1.4 functions used by the GLMesh class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLMesh`
class to render :class:`.Mesh` overlays in an OpenGL 1.4 compatible
manner.
"""


import fsleyes.gl.shaders as shaders


def compileShaders(self):
    """Loads the ``glmesh`` vertex and fragment shader program source,
    and creates :class:`.ARBPShader` instance(s).
    """

    textures = {
        'cmap'    : 0,
        'negCmap' : 1
    }

    if self.threedee:

        flatVertSrc = shaders.getVertexShader(  'glmesh_3d_flat')
        flatFragSrc = shaders.getFragmentShader('glmesh_3d_flat')
        dataVertSrc = shaders.getVertexShader(  'glmesh_3d_data')
        dataFragSrc = shaders.getFragmentShader('glmesh_3d_data')

        self.flatShader = shaders.ARBPShader(flatVertSrc,
                                             flatFragSrc)
        self.dataShader = shaders.ARBPShader(dataVertSrc,
                                             dataFragSrc,
                                             textures)

    else:

        flatVertSrc    = shaders.getVertexShader(  'glmesh_2d_flat')
        flatFragSrc    = shaders.getFragmentShader('glmesh_2d_flat')
        dataVertSrc    = shaders.getVertexShader(  'glmesh_2d_data')
        dataFragSrc    = shaders.getFragmentShader('glmesh_2d_data')
        xsectcpVertSrc = shaders.getVertexShader(  'glmesh_2d_crosssection_clipplane')
        xsectcpFragSrc = shaders.getFragmentShader('glmesh_2d_crosssection_clipplane')
        xsectblVertSrc = shaders.getVertexShader(  'glmesh_2d_crosssection_blit')
        xsectblFragSrc = shaders.getFragmentShader('glmesh_2d_crosssection_blit')

        self.flatShader    = shaders.ARBPShader(flatVertSrc,    flatFragSrc)
        self.dataShader    = shaders.ARBPShader(dataVertSrc,    dataFragSrc)
        self.xsectcpShader = shaders.ARBPShader(xsectcpVertSrc, xsectcpFragSrc)
        self.xsectblShader = shaders.ARBPShader(xsectblVertSrc, xsectblFragSrc)


def updateShaderState(self, **kwargs):
    """Updates the state of the vertex/fragment shaders according to the
    current :class:`.MeshOpts` configuration. This involves setting the
    parameter values used by the shaders.
    """
    dopts      = self.opts
    dshader    = self.dataShader
    fshader    = self.flatShader
    xscpshader = self.xsectcpShader
    xsblshader = self.xsectblShader

    settings = [-1 if     kwargs['useNegCmap'] else 1,
                -1 if     dopts.invertClipping else 1,
                -1 if not dopts.discardClipped else 1,
                0]

    clipping = [dopts.clippingRange.xlo, dopts.clippingRange.xhi, 0, 0]
    modulate = [-1 if not dopts.modulateAlpha  else 1,
                kwargs['modScale'],
                kwargs['modOffset'],
                0]

    with dshader.loaded():
        dshader.setFragParam('settings',   settings)
        dshader.setFragParam('clipping',   clipping)
        dshader.setFragParam('modulate',   modulate)
        dshader.setFragParam('flatColour', kwargs['flatColour'])
        dshader.setFragParam('cmapXform',  kwargs['cmapXform'])

        if self.threedee:
            dshader.setIndices(self.indices)
            dshader.setAtt('vertex', self.vertices)
            dshader.setAtt('normal', self.normals)
            vdata = self.getVertexData('vertex')
            mdata = self.getVertexData('modulate')

            # if modulate data is not set,
            # we use the vertex data
            if mdata is None:
                mdata = vdata

            if vdata is not None: dshader.setAtt('vertexData',   vdata)
            if mdata is not None: dshader.setAtt('modulateData', mdata)

    with fshader.loaded():
        fshader.setFragParam('colour', kwargs['flatColour'])
        if self.threedee:
            fshader.setAtt('vertex', self.vertices)
            fshader.setAtt('normal', self.normals)
            fshader.setIndices(self.indices)

    if not self.threedee:
        with xscpshader.loaded():
            xscpshader.setAtt('vertex', self.vertices)
            xscpshader.setIndices(      self.indices)
        with xsblshader.loaded():
            xsblshader.set('colour', kwargs['flatColour'])



#!/usr/bin/env python
#
# gllabel.py - The GLLabel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLLabel` class, which implements
functionality to render an :class:`.Image` overlay as a label/atlas image.
"""


import OpenGL.GL                 as gl

import fsl.utils.idle            as idle
import fsl.transform.affine      as affine
import fsleyes.gl                as fslgl
import fsleyes.gl.routines       as glroutines
import fsleyes.gl.shaders.filter as glfilter
from . import resources          as glresources
from . import                       glimageobject
from . import                       textures


class GLLabel(glimageobject.GLImageObject):
    """The ``GLLabel`` class is a :class:`.GLImageObject` which encapsulates
    the logic required to render an :class:`.Image` overlay as a label image.
    Within the image, each contiguous region with the same label value is
    rendered in the same colour. Regions may be shown either with a filled
    colour, or with a border around them.

    When created, a ``GLLabel`` instance assumes that the provided
    :class:`.Image` instance has a :attr:`.Display.overlayType` of ``label``,
    and that its associated :class:`.Display` instance contains a
    :class:`.LabelOpts` instance, containing label-specific display
    properties.

    An :class:`.ImageTexture` is used to store the :class:`.Image` data, and
    a :class:`.LookupTableTexture` used to store the :class:`.LookupTable`
    (defined by the :attr:`.LabelOpts.lut` property). OpenGL version-specific
    modules (:mod:`.gl14.gllabel_funcs` and :mod:`.gl21.gllabel_funcs`) are
    used to configure the vertex/fragment shader programs used for rendering.

    The ``GLLabel`` class is modelled upon the :class:`.GLVolume` class, and
    the version specific modules for the ``GLLabel`` class must provide the
    same set of functions that are required by the ``GLVolume`` class.
    """


    def __init__(self, image, overlayList, displayCtx, threedee):
        """Create a ``GLLabel``.

        :arg image:       The :class:`.Image` instance.
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
        :arg threedee:    2D or 3D rendering
        """

        glimageobject.GLImageObject.__init__(self,
                                             image,
                                             overlayList,
                                             displayCtx,
                                             threedee)

        # The shader attribute will be created
        # by the gllabel_funcs module, and the
        # imageTexture by the refreshTexture
        # method
        self.shader       = None
        self.imageTexture = None
        self.lutTexture   = textures.LookupTableTexture(
            '{}_lut'.format(self.name))
        self.edgeFilter   = glfilter.Filter('edge', texture=0)
        self.renderTexture = textures.RenderTexture(
            self.name, interp=gl.GL_LINEAR, rttype='c')

        self.__lut = self.opts.lut

        self.addListeners()
        self.registerLut()
        self.refreshLutTexture()
        self.refreshImageTexture()

        def init():
            if not self.destroyed:
                fslgl.gllabel_funcs.init(self)
                self.notify()

        idle.idleWhen(init, self.textureReady)


    def destroy(self):
        """Must be called when this ``GLLabel`` is no longer needed. Destroys
        the :class:`.ImageTexture` and :class:`.LookupTableTexture`.
        """

        self.edgeFilter.destroy()
        self.renderTexture.destroy()
        self.imageTexture.deregister(self.name)
        glresources.delete(self.imageTexture.name)
        self.lutTexture.destroy()

        self.removeListeners()
        fslgl.gllabel_funcs.destroy(self)
        glimageobject.GLImageObject.destroy(self)


    def ready(self):
        """Returns ``True`` if this ``GLLabel`` is ready to be drawn, ``False``
        otherwise.
        """
        return self.shader is not None and self.textureReady()


    def textureReady(self):
        """Returns ``True`` if the ``imageTexture`` is ready to be used,
        ``False`` otherwise.
        """
        return self.imageTexture is not None and self.imageTexture.ready()


    def updateShaderState(self, *args, **kwargs):
        """Calls :func:`.gl14.gllabel_funcs.updateShaderState` or
        :func:`.gl21.gllabel_funcs.updateShaderState`, and
        :meth:`.Notifier.notify`. Uses :func:`.idle.idleWhen` to ensure that
        they don't get called until :meth:`ready` returns ``True``.
        """
        alwaysNotify = kwargs.pop('alwaysNotify', None)

        def func():
            if fslgl.gllabel_funcs.updateShaderState(self) or alwaysNotify:
                self.notify()

        idle.idleWhen(func,
                      self.ready,
                      name=self.name,
                      skipIfQueued=True)


    def addListeners(self):
        """Called by :meth:`__init__`. Adds listeners to several properties of
        the :class:`.Display` and :class:`.LabelOpts` instances, so the OpenGL
        representation can be updated when they change.
        """

        display = self.display
        opts    = self.opts
        name    = self.name

        display.addListener('alpha',        name, self.__colourPropChanged)
        display.addListener('brightness',   name, self.__colourPropChanged)
        display.addListener('contrast',     name, self.__colourPropChanged)
        opts   .addListener('outline',      name, self.notify)
        opts   .addListener('outlineWidth', name, self.notify)
        opts   .addListener('lut',          name, self.__lutChanged)
        opts   .addListener('volume',       name, self.__imagePropChanged)
        opts   .addListener('transform',    name, self.notify)

        # See comment in GLVolume.addDisplayListeners about this
        self.__syncListenersRegistered = opts.getParent() is not None

        if self.__syncListenersRegistered:
            opts.addSyncChangeListener(
                'volume',     name, self.__imageSyncChanged)


    def removeListeners(self):
        """Called by :meth:`destroy`. Removes all of the listeners that were
        added by :meth:`addListeners`.
        """

        display = self.display
        opts    = self.opts
        name    = self.name

        display.removeListener('alpha',        name)
        display.removeListener('brightness',   name)
        display.removeListener('contrast',     name)
        opts   .removeListener('outline',      name)
        opts   .removeListener('outlineWidth', name)
        opts   .removeListener('lut',          name)
        opts   .removeListener('volume',       name)
        opts   .removeListener('transform',    name)

        if self.__syncListenersRegistered:
            opts.removeSyncChangeListener('volume', name)


    def refreshImageTexture(self):
        """Makes sure that the :class:`.ImageTexture`, used to store the
        :class:`.Image` data, is up to date.
        """

        opts     = self.opts
        texName  = '{}_{}' .format(type(self).__name__, id(self.image))

        unsynced = (opts.getParent() is None or
                    not opts.isSyncedToParent('volume'))

        if unsynced:
            texName = '{}_unsync_{}'.format(texName, id(opts))

        if self.imageTexture is not None:

            if self.imageTexture.name == texName:
                return None

            self.imageTexture.deregister(self.name)
            glresources.delete(self.imageTexture.name)

        self.imageTexture = glresources.get(
            texName,
            textures.ImageTexture,
            texName,
            self.image,
            notify=False,
            volume=opts.index()[3:])

        self.imageTexture.register(self.name, self.__imageTextureChanged)


    def refreshLutTexture(self, *a):
        """Refreshes the :class:`.LookupTableTexture` which stores the
        :class:`.LookupTable` used to colour the overlay.
        """

        display = self.display
        opts    = self.opts

        self.lutTexture.set(alpha=display.alpha           / 100.0,
                            brightness=display.brightness / 100.0,
                            contrast=display.contrast     / 100.0,
                            lut=opts.lut)


    def registerLut(self):
        """Registers a listener on the current :class:`.LookupTable` instance.
        """
        opts = self.opts

        if self.__lut is not None:
            for topic in ['label', 'added', 'removed']:
                self.__lut.deregister(self.name, topic)

        self.__lut = opts.lut

        if self.__lut is not None:
            for topic in ['label', 'added', 'removed']:
                self.__lut.register(self.name, self.__colourPropChanged, topic)


    def preDraw(self):
        """Binds the :class:`.ImageTexture` and :class:`.LookupTableTexture`.
        """
        self.imageTexture.bindTexture(gl.GL_TEXTURE0)
        self.lutTexture  .bindTexture(gl.GL_TEXTURE1)


    def draw2D(self, canvas, zpos, axes, xform=None):
        """Calls the version-dependent ``draw2D`` function. """

        opts       = self.opts
        outline    = opts.outline
        owidth     = float(opts.outlineWidth)
        rtex       = self.renderTexture
        w, h       = canvas.GetSize()
        bbox       = canvas.viewport
        projmat    = canvas.projectionMatrix
        viewmat    = canvas.viewMatrix
        lo         = [ax[0] for ax in bbox]
        hi         = [ax[1] for ax in bbox]
        xax        = axes[0]
        yax        = axes[1]
        xmin, xmax = bbox[xax]
        ymin, ymax = bbox[yax]
        offsets    = [owidth / w, owidth / h]

        rtex.shape = w, h

        # draw the labels to the offscreen texture
        with glroutines.disabled(gl.GL_BLEND), \
             rtex.target(xax, yax, lo, hi):
            gl.glClearColor(0, 0, 0, 0)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT)
            fslgl.gllabel_funcs.draw2D(self, rtex, zpos, axes, xform)

        # run the offscreen texture through the
        # edge filter, drawing the result to
        # the canvas
        if xform is None: xform = affine.concat(projmat, viewmat)
        else:             xform = affine.concat(projmat, viewmat, xform)

        self.edgeFilter.set(offsets=offsets, outline=outline)
        self.edgeFilter.apply(rtex, zpos, xmin, xmax,
                              ymin, ymax, xax, yax, xform)


    def draw3D(self, canvas, xform=None):
        """Not implemented."""


    def drawAll(self, canvas, axes, zposes, xforms):
        """Calls the version-dependent ``drawAll`` function. """

        opts       = self.opts
        outline    = opts.outline
        owidth     = float(opts.outlineWidth)
        rtex       = self.renderTexture
        w, h       = canvas.GetSize()
        bbox       = canvas.viewport
        projmat    = canvas.projectionMatrix
        viewmat    = canvas.viewMatrix
        lo         = [ax[0] for ax in bbox]
        hi         = [ax[1] for ax in bbox]
        xax        = axes[0]
        yax        = axes[1]
        xmin, xmax = bbox[xax]
        ymin, ymax = bbox[yax]
        offsets    = [owidth / w, owidth / h]

        rtex.shape = w, h

        # draw all slices to the offscreen texture
        with glroutines.disabled(gl.GL_BLEND), \
             rtex.target(xax, yax, lo, hi):
            gl.glClearColor(0, 0, 0, 0)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT)
            fslgl.gllabel_funcs.drawAll(self, rtex, axes, zposes, xforms)

        xform = affine.concat(projmat, viewmat)

        # run it through the edge filter
        self.edgeFilter.set(offsets=offsets, outline=outline)
        self.edgeFilter.apply(rtex, max(zposes), xmin, xmax,
                              ymin, ymax, xax, yax, xform)


    def postDraw(self):
        """Unbinds the ``ImageTexture`` and ``LookupTableTexture``.
        """
        self.imageTexture.unbindTexture()
        self.lutTexture  .unbindTexture()


    def __lutChanged(self, *a):
        """Called when the :attr:`.LabelOpts.lut` property changes. Re-creates
        the :class:`.LookupTableTexture`.
        """

        self.registerLut()
        self.refreshLutTexture()
        self.updateShaderState(alwaysNotify=True)


    def __colourPropChanged(self, *a):
        """Called when a :class:`.Display` property changes (e.g. ``alpha``).
        Refreshes the LUT texture.
        """
        self.refreshLutTexture()
        self.updateShaderState(alwaysNotify=True)


    def __imagePropChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.volume` property changes. Updates
        the ``imageTexture`` and calls :meth:`updateShaderState`.
        """
        opts = self.opts

        self.imageTexture.set(volume=opts.index()[3:])
        self.updateShaderState(alwaysNotify=True)


    def __imageSyncChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.volume` property is synchronised
        or un-synchronised. Calls :meth:`refreshImageTexture` and
        :meth:`updateShaderState`.
        """
        self.refreshImageTexture()
        self.updateShaderState(alwaysNotify=True)


    def __imageTextureChanged(self, *a):
        """Called when the :class:`.ImageTexture` containing the image data
        changes. Calls :meth:`updateShaderState`.
        """
        self.updateShaderState(alwaysNotify=True)



#!/usr/bin/env python
#
# gllabel_funcs.py - OpenGL 2.1 functions used by the GLLabel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions which are used by the :class:`.GLLabel`
class to render :class:`.Image` overlays in an OpenGL 2.1 compatible manner.

Rendering of a ``GLLabel`` is very similar to that of a ``GLVolume``, with the
exception that a different fragment shader (``gllabel``) is used. The
``preDraw``, ``draw``, ``drawAll`` and ``postDraw`` functions defined in the
:mod:`.gl21.glvolume_funcs` are re-used by this module.
"""


import numpy              as np

import fsleyes.gl.shaders as shaders
from . import                glvolume_funcs


def init(self):
    """Calls the :func:`compileShaders` and :func:`updateShaderState`
    functions.
    """
    self.shader = None

    compileShaders(   self)
    updateShaderState(self)


def destroy(self):
    """Destroys the shader programs. """
    if self.shader is not None:
        self.shader.destroy()
    self.shader = None


def compileShaders(self):
    """Loads the vertex/fragment shader source code, and creates a
    :class:`.GLSLShader` program.
    """

    if self.shader is not None:
        self.shader.destroy()

    vertSrc = shaders.getVertexShader(  'glvolume')
    fragSrc = shaders.getFragmentShader('gllabel')

    self.shader = shaders.GLSLShader(vertSrc, fragSrc)


def updateShaderState(self):
    """Updates all shader program variables. """

    if not self.ready():
        return

    opts   = self.opts
    shader = self.shader

    imageShape = np.array(self.image.shape[:3])
    vvx        = self.imageTexture.voxValXform

    with shader.loaded():
        changed  = False
        changed |= shader.set('numLabels',    opts.lut.max() + 1)
        changed |= shader.set('imageShape',   imageShape)
        changed |= shader.set('voxValXform',  vvx)
        changed |= shader.set('imageTexture', 0)
        changed |= shader.set('lutTexture',   1)

    return changed


def draw2D(self, *args, **kwargs):
    """Draws the label overlay in 2D. See :meth:`.GLObject.draw2D`."""
    with self.shader.loaded():
        glvolume_funcs.draw2D(self, *args, **kwargs)


def drawAll(self, *args, **kwargs):
    """Draws the label overlay in 2D. See :meth:`.GLObject.draw2D`."""
    with self.shader.loaded():
        glvolume_funcs.drawAll(self, *args, **kwargs)
